import os
from datetime import datetime, timedelta
from pathlib import Path
from feathr.client import FeathrClient

from pyspark.sql import DataFrame
from pyspark.sql.functions import col

from feathr import (BackfillTime, MaterializationSettings)
from feathr import Feature
from feathr import FeatureAnchor
from feathr import FeatureQuery
from feathr import HdfsSource, SnowflakeSource
from feathr import ObservationSettings
from feathr import RedisSink
from feathr import STRING, FLOAT, INT32, ValueType
from feathr import TypedKey
from feathr import WindowAggTransformation
from feathr.utils.job_utils import get_result_df
from test_fixture import (snowflake_test_setup, get_online_test_table_name, basic_test_setup)
from test_utils.constants import Constants


def trip_distance_preprocessing(df: DataFrame):
    df = df.withColumn("trip_distance", df.trip_distance.cast('double') - 90000)
    df = df.withColumn("fare_amount", df.fare_amount.cast('double') - 90000)

    return df

def add_new_dropoff_and_fare_amount_column(df: DataFrame):
    df = df.withColumn("new_lpep_dropoff_datetime", col("lpep_dropoff_datetime"))
    df = df.withColumn("new_fare_amount", col("fare_amount") + 1000000)
    return df

def add_new_fare_amount(df: DataFrame) -> DataFrame:
    df = df.withColumn("fare_amount_new", col("fare_amount") + 8000000)

    return df

def add_new_surcharge_amount_and_pickup_column(df: DataFrame) -> DataFrame:
    df = df.withColumn("new_improvement_surcharge", col("improvement_surcharge") + 1000000)
    df = df.withColumn("new_tip_amount", col("tip_amount") + 1000000)
    df = df.withColumn("new_lpep_pickup_datetime", col("lpep_pickup_datetime"))

    return df

def add_old_lpep_dropoff_datetime(df: DataFrame) -> DataFrame:
    df = df.withColumn("old_lpep_dropoff_datetime", col("lpep_dropoff_datetime"))

    return df

def feathr_udf_day_calc(df: DataFrame) -> DataFrame:
    df = df.withColumn("f_day_of_week", dayofweek("lpep_dropoff_datetime"))
    df = df.withColumn("f_day_of_year", dayofyear("lpep_dropoff_datetime"))
    return df

def test_non_swa_feature_gen_with_offline_preprocessing():
    """
    Test non-SWA feature gen with preprocessing
    """
    test_workspace_dir = Path(__file__).parent.resolve() / "test_user_workspace"

    client:FeathrClient = basic_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))

    batch_source = HdfsSource(name="nycTaxiBatchSource_add_new_fare_amount",
                              path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv",
                              preprocessing=add_new_fare_amount,
                              event_timestamp_column="lpep_dropoff_datetime",
                              timestamp_format="yyyy-MM-dd HH:mm:ss")

    pickup_time_as_id = TypedKey(key_column="lpep_pickup_datetime",
                                 key_column_type=ValueType.INT32,
                                 description="location id in NYC",
                                 full_name="nyc_taxi.location_id")

    features = [
        Feature(name="f_is_long_trip_distance",
                key=pickup_time_as_id,
                feature_type=FLOAT,
                transform="fare_amount_new"),
        Feature(name="f_day_of_week",
                key=pickup_time_as_id,
                feature_type=INT32,
                transform="dayofweek(lpep_dropoff_datetime)"),
    ]

    regular_anchor = FeatureAnchor(name="request_features_add_new_fare_amount",
                                   source=batch_source,
                                   features=features,
                                   )

    client.build_features(anchor_list=[regular_anchor])

    online_test_table = get_online_test_table_name('nycTaxiCITableOfflineProcessing')

    backfill_time = BackfillTime(start=datetime(
        2020, 5, 20), end=datetime(2020, 5, 20), step=timedelta(days=1))
    redisSink = RedisSink(table_name=online_test_table)
    settings = MaterializationSettings(name="py_udf",
                                       sinks=[redisSink],
                                       feature_names=[
                                           "f_is_long_trip_distance",
                                           "f_day_of_week"
                                       ],
                                       backfill_time=backfill_time)
    client.materialize_features(settings, allow_materialize_non_agg_feature=True)
    # just assume the job is successful without validating the actual result in Redis. Might need to consolidate
    # this part with the test_feathr_online_store test case
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)

    res = client.get_online_features(online_test_table, '2020-04-01 07:21:51', [
        'f_is_long_trip_distance', 'f_day_of_week'])
    assert res == [8000006.0, 4]
    
    client._clean_test_data(online_test_table)


def test_feature_swa_feature_gen_with_preprocessing():
    """
    Test SWA feature gen with preprocessing.
    """
    test_workspace_dir = Path(__file__).parent.resolve() / "test_user_workspace"
    client = basic_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))

    batch_source = HdfsSource(name="nycTaxiBatchSource",
                              path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv",
                              preprocessing=add_new_dropoff_and_fare_amount_column,
                              event_timestamp_column="new_lpep_dropoff_datetime",
                              timestamp_format="yyyy-MM-dd HH:mm:ss")

    location_id = TypedKey(key_column="DOLocationID",
                           key_column_type=ValueType.INT32,
                           description="location id in NYC",
                           full_name="nyc_taxi.location_id")


    agg_features = [Feature(name="f_location_avg_fare",
                            key=location_id,
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="new_fare_amount",
                                                              agg_func="AVG",
                                                              window="90d")),
                    Feature(name="f_location_max_fare",
                            key=location_id,
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="new_fare_amount",
                                                              agg_func="MAX",
                                                              window="90d"))
                    ]

    agg_anchor = FeatureAnchor(name="aggregationFeatures",
                               source=batch_source,
                               features=agg_features)

    client.build_features(anchor_list=[agg_anchor])

    online_test_table = get_online_test_table_name('nycTaxiCITableSWAFeatureMaterialization')

    backfill_time = BackfillTime(start=datetime(
        2020, 5, 20), end=datetime(2020, 5, 20), step=timedelta(days=1))
    redisSink = RedisSink(table_name=online_test_table)
    settings = MaterializationSettings(name="py_udf",
                                       sinks=[redisSink],
                                       feature_names=[
                                           "f_location_avg_fare",
                                           "f_location_max_fare",
                                       ],
                                       backfill_time=backfill_time)
    client.materialize_features(settings)
    # just assume the job is successful without validating the actual result in Redis. Might need to consolidate
    # this part with the test_feathr_online_store test case
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)

    res = client.get_online_features(online_test_table, '265', ['f_location_avg_fare', 'f_location_max_fare'])
    assert res == [1000041.625, 1000100.0]
    
    client._clean_test_data(online_test_table)


def test_feathr_get_offline_features_hdfs_source():
    """
    Test get offline features for blob storage
    """
    test_workspace_dir = Path(__file__).parent.resolve() / "test_user_workspace"

    client = basic_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))

    batch_source1 = HdfsSource(name="nycTaxiBatchSource_add_new_dropoff_and_fare_amount_column",
                              path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv",
                              preprocessing=add_new_dropoff_and_fare_amount_column,
                              event_timestamp_column="new_lpep_dropoff_datetime",
                              # event_timestamp_column="lpep_dropoff_datetime",
                              timestamp_format="yyyy-MM-dd HH:mm:ss")

    batch_source2 = HdfsSource(name="nycTaxiBatchSource_add_new_fare_amount",
                              path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv",
                              preprocessing=add_new_fare_amount,
                              event_timestamp_column="lpep_dropoff_datetime",
                              timestamp_format="yyyy-MM-dd HH:mm:ss")


    location_id = TypedKey(key_column="DOLocationID",
                           key_column_type=ValueType.INT32,
                           description="location id in NYC",
                           full_name="nyc_taxi.location_id")


    agg_features = [Feature(name="f_location_avg_fare",
                            key=location_id,
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="new_fare_amount",
                                                              agg_func="SUM",
                                                              window="90d")),
                    Feature(name="f_location_max_fare",
                            key=location_id,
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="new_fare_amount",
                                                              agg_func="MAX",
                                                              window="90d"))
                    ]

    agg_anchor = FeatureAnchor(name="aggregationFeatures",
                               source=batch_source1,
                               features=agg_features,
                               )

    pickup_time_as_id = TypedKey(key_column="lpep_pickup_datetime",
                                 key_column_type=ValueType.INT32,
                                 description="location id in NYC",
                                 full_name="nyc_taxi.location_id")

    features = [
        Feature(name="f_is_long_trip_distance",
                key=pickup_time_as_id,
                feature_type=FLOAT,
                transform="fare_amount_new"),
        Feature(name="f_day_of_week",
                key=pickup_time_as_id,
                feature_type=INT32,
                transform="dayofweek(lpep_dropoff_datetime)"),
    ]

    regular_anchor = FeatureAnchor(name="regular_anchor",
                                   source=batch_source2,
                                   features=features,
                                   )

    client.build_features(anchor_list=[agg_anchor, regular_anchor])

    feature_query = [FeatureQuery(
        feature_list=["f_is_long_trip_distance", "f_day_of_week"], key=pickup_time_as_id),
        FeatureQuery(
            feature_list=["f_location_avg_fare", "f_location_max_fare"], key=location_id)
    ]

    settings = ObservationSettings(
        observation_path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv",
        event_timestamp_column="lpep_dropoff_datetime",
        timestamp_format="yyyy-MM-dd HH:mm:ss")

    now = datetime.now()
    # set output folder based on different runtime
    if client.spark_runtime == 'databricks':
        output_path = ''.join(['dbfs:/feathrazure_cijob','_', str(now.minute), '_', str(now.second), ".avro"])
    else:
        output_path = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/demo_data/output','_', str(now.minute), '_', str(now.second), ".avro"])


    client.get_offline_features(observation_settings=settings,
                                feature_query=feature_query,
                                output_path=output_path)

    # assuming the job can successfully run; otherwise it will throw exception
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)

    # download result and just assert the returned result is not empty
    res_df = get_result_df(client)
    assert res_df.shape[0] > 0


def test_get_offline_feature_two_swa_with_diff_preprocessing():
    """
    Test get offline features for two SWA anchors with different preprocessing and different sources.
    """
    test_workspace_dir = Path(__file__).parent.resolve() / "test_user_workspace"

    client = basic_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))

    swa_source_1 = HdfsSource(name="nycTaxiBatchSource1",
                               path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv",
                               preprocessing=add_new_dropoff_and_fare_amount_column,
                               event_timestamp_column="new_lpep_dropoff_datetime",
                               timestamp_format="yyyy-MM-dd HH:mm:ss")

    location_id = TypedKey(key_column="DOLocationID",
                           key_column_type=ValueType.INT32,
                           description="location id in NYC",
                           full_name="nyc_taxi.location_id")


    agg_features1 = [Feature(name="f_location_avg_fare",
                            key=location_id,
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="new_fare_amount",
                                                              agg_func="SUM",
                                                              window="90d")),
                    Feature(name="f_location_max_fare",
                            key=location_id,
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="new_fare_amount",
                                                              agg_func="MAX",
                                                              window="90d"))
                    ]

    agg_anchor1 = FeatureAnchor(name="aggregationFeatures1",
                               source=swa_source_1,
                               features=agg_features1,
                               )


    swa_source_2 = HdfsSource(name="nycTaxiBatchSource2",
                              path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv",
                              preprocessing=add_new_surcharge_amount_and_pickup_column,
                              event_timestamp_column="new_lpep_pickup_datetime",
                              timestamp_format="yyyy-MM-dd HH:mm:ss")

    agg_features2 = [Feature(name="f_location_new_tip_amount",
                             key=location_id,
                             feature_type=FLOAT,
                             transform=WindowAggTransformation(agg_expr="new_tip_amount",
                                                               agg_func="SUM",
                                                               window="90d")),
                     Feature(name="f_location_max_improvement_surcharge",
                             key=location_id,
                             feature_type=FLOAT,
                             transform=WindowAggTransformation(agg_expr="new_improvement_surcharge",
                                                               agg_func="SUM",
                                                               window="90d"))
                     ]
    agg_anchor2 = FeatureAnchor(name="aggregationFeatures2",
                               source=swa_source_2,
                               features=agg_features2,
                               )

    swa_source_3 = HdfsSource(name="nycTaxiBatchSource3",
                              path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04_old.csv",
                              preprocessing=add_old_lpep_dropoff_datetime,
                              event_timestamp_column="old_lpep_dropoff_datetime",
                              timestamp_format="yyyy-MM-dd HH:mm:ss")

    agg_features3 = [Feature(name="f_location_old_tip_amount",
                             key=location_id,
                             feature_type=FLOAT,
                             transform=WindowAggTransformation(agg_expr="cast_double(old_tip_amount)",
                                                               agg_func="SUM",
                                                               window="90d"))
                     ]
    agg_anchor3 = FeatureAnchor(name="aggregationFeatures3",
                                source=swa_source_3,
                                features=agg_features3,
                                )

    client.build_features(anchor_list=[agg_anchor1, agg_anchor2, agg_anchor3])

    feature_query = [
        FeatureQuery(feature_list=["f_location_new_tip_amount", "f_location_max_improvement_surcharge"], key=location_id),
        FeatureQuery(
            feature_list=["f_location_avg_fare", "f_location_max_fare"], key=location_id),
        FeatureQuery(
            feature_list=["f_location_old_tip_amount"], key=location_id)
    ]

    settings = ObservationSettings(
        observation_path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv",
        event_timestamp_column="lpep_dropoff_datetime",
        timestamp_format="yyyy-MM-dd HH:mm:ss")

    now = datetime.now()
    # set output folder based on different runtime
    if client.spark_runtime == 'databricks':
        output_path = ''.join(['dbfs:/feathrazure_cijob','_', str(now.minute), '_', str(now.second), ".avro"])
    else:
        output_path = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/demo_data/output','_', str(now.minute), '_', str(now.second), ".avro"])

    client.get_offline_features(observation_settings=settings,
                                feature_query=feature_query,
                                output_path=output_path)

    # assuming the job can successfully run; otherwise it will throw exception
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)
    res_df = get_result_df(client)

    # download result and just assert the returned result is not empty
    assert res_df.shape[0] > 0


def snowflake_preprocessing(df: DataFrame) -> DataFrame:
    df = df.withColumn("NEW_CC_DIVISION_NAME", concat(col("CC_DIVISION_NAME"), lit("0000"), col("CC_DIVISION_NAME")))
    df = df.withColumn("NEW_CC_ZIP", concat(col("CC_ZIP"), lit("____"), col("CC_ZIP")))
    return df


def test_feathr_get_offline_features_from_snowflake():
    """
    Test get_offline_features() can get feature data from Snowflake source correctly.
    """
    test_workspace_dir = Path(__file__).parent.resolve() / "test_user_workspace"
    client = snowflake_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))
    batch_source = SnowflakeSource(name="nycTaxiBatchSource",
                                    database="SNOWFLAKE_SAMPLE_DATA",
                                    schema="TPCDS_SF10TCL",
                                    dbtable="CALL_CENTER",
                                    preprocessing=snowflake_preprocessing,
                                                            event_timestamp_column="lpep_dropoff_datetime",
                                                            timestamp_format="yyyy-MM-dd HH:mm:ss")
    call_sk_id = TypedKey(key_column="CC_CALL_CENTER_SK",
                          key_column_type=ValueType.STRING,
                          description="call center sk",
                          full_name="snowflake.CC_CALL_CENTER_SK")

    features = [
        Feature(name="f_snowflake_call_center_division_name_with_preprocessing",
                key=call_sk_id,
                feature_type=STRING,
                transform="NEW_CC_DIVISION_NAME"),
        Feature(name="f_snowflake_call_center_zipcode_with_preprocessing",
                key=call_sk_id,
                feature_type=STRING,
                transform="NEW_CC_ZIP"),
    ]

    feature_anchor = FeatureAnchor(name="snowflake_features",
                                   source=batch_source,
                                   features=features,
                                   )
    client.build_features(anchor_list=[feature_anchor])

    feature_query = FeatureQuery(
        feature_list=['f_snowflake_call_center_division_name_with_preprocessing', 'f_snowflake_call_center_zipcode_with_preprocessing'],
        key=call_sk_id)

    observation_path = client.get_snowflake_path(database="SNOWFLAKE_SAMPLE_DATA", schema="TPCDS_SF10TCL", dbtable="CALL_CENTER")
    settings = ObservationSettings(
        observation_path=observation_path)

    now = datetime.now()
    # set output folder based on different runtime
    if client.spark_runtime == 'databricks':
        output_path = ''.join(['dbfs:/feathrazure_cijob_snowflake', '_', str(now.minute), '_', str(now.second), ".avro"])
    else:
        output_path = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/demo_data/snowflake_output','_', str(now.minute), '_', str(now.second), ".avro"])

    client.get_offline_features(observation_settings=settings,
                                feature_query=feature_query,
                                output_path=output_path)

    # assuming the job can successfully run; otherwise it will throw exception
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)

    res = get_result_df(client)
    # just assume there are results.
    assert res.shape[0] > 1



import os
from datetime import datetime, timedelta
from pathlib import Path
from feathr import (BOOLEAN, FLOAT, INPUT_CONTEXT, INT32,
                    DerivedFeature, Feature, FeatureAnchor,
                    TypedKey, ValueType, WindowAggTransformation)

import pytest
from feathr import FeathrClient
from feathr import FeatureQuery
from feathr import ObservationSettings
from feathr import TypedKey
from feathr import ValueType
from feathr.definition.materialization_settings import BackfillTime, MaterializationSettings
from feathr.definition.sink import HdfsSink
from feathr.utils.job_utils import get_result_df
from test_utils.constants import Constants
from feathr.definition.source import SparkSqlSource


def test_feathr_spark_sql_query_source():
    test_workspace_dir = Path(
        __file__).parent.resolve() / "test_user_workspace"
    config_path = os.path.join(test_workspace_dir, "feathr_config.yaml")

    _get_offline_features(config_path, _sql_query_source())
    _get_offline_features(config_path, _sql_table_source())
    _materialize_to_offline(config_path, _sql_query_source())


def _get_offline_features(config_path: str, sql_source: SparkSqlSource):
    client: FeathrClient = _spark_sql_test_setup(config_path, sql_source)
    location_id = TypedKey(key_column="DOLocationID",
                           key_column_type=ValueType.INT32,
                           description="location id in NYC",
                           full_name="nyc_taxi.location_id")

    feature_query = FeatureQuery(
        feature_list=["f_location_avg_fare"], key=location_id)
    settings = ObservationSettings(
        observation_path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv",
        event_timestamp_column="lpep_dropoff_datetime",
        timestamp_format="yyyy-MM-dd HH:mm:ss")

    now = datetime.now()
    if client.spark_runtime == 'databricks':
        output_path = ''.join(['dbfs:/feathrazure_cijob_materialize_offline_','_', str(now.minute), '_', str(now.second), ""])
    else:
        output_path = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/demo_data/feathrazure_cijob_materialize_offline_','_', str(now.minute), '_', str(now.second), ""])
    client.get_offline_features(observation_settings=settings,
                                feature_query=feature_query,
                                output_path=output_path)

    # assuming the job can successfully run; otherwise it will throw exception
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)
    return


def _materialize_to_offline(config_path: str, sql_source: SparkSqlSource):
    client: FeathrClient = _spark_sql_test_setup(config_path, sql_source)
    backfill_time = BackfillTime(start=datetime(
        2020, 5, 20), end=datetime(2020, 5, 20), step=timedelta(days=1))

    now = datetime.now()
    if client.spark_runtime == 'databricks':
        output_path = ''.join(['dbfs:/feathrazure_cijob_materialize_offline_sparksql',
                              '_', str(now.minute), '_', str(now.second), ""])
    else:
        output_path = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/demo_data/feathrazure_cijob_materialize_offline_sparksql',
                              '_', str(now.minute), '_', str(now.second), ""])
    offline_sink = HdfsSink(output_path=output_path)
    settings = MaterializationSettings("nycTaxiTable",
                                       sinks=[offline_sink],
                                       feature_names=[
                                           "f_location_avg_fare", "f_location_max_fare"],
                                       backfill_time=backfill_time)
    client.materialize_features(settings)
    # assuming the job can successfully run; otherwise it will throw exception
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)

    # download result and just assert the returned result is not empty
    # by default, it will write to a folder appended with date
    res_df = get_result_df(
        client, "avro", output_path + "/df0/daily/2020/05/20")
    assert res_df.shape[0] > 0


def _spark_sql_test_setup(config_path: str, sql_source: SparkSqlSource):
    client = FeathrClient(config_path=config_path)

    f_trip_distance = Feature(name="f_trip_distance",
                              feature_type=FLOAT, transform="trip_distance")
    f_trip_time_duration = Feature(name="f_trip_time_duration",
                                   feature_type=INT32,
                                   transform="(to_unix_timestamp(lpep_dropoff_datetime) - to_unix_timestamp(lpep_pickup_datetime))/60")

    features = [
        f_trip_distance,
        f_trip_time_duration,
        Feature(name="f_is_long_trip_distance",
                feature_type=BOOLEAN,
                transform="cast_float(trip_distance)>30"),
        Feature(name="f_day_of_week",
                feature_type=INT32,
                transform="dayofweek(lpep_dropoff_datetime)"),
    ]

    request_anchor = FeatureAnchor(name="request_features",
                                   source=INPUT_CONTEXT,
                                   features=features)

    f_trip_time_distance = DerivedFeature(name="f_trip_time_distance",
                                          feature_type=FLOAT,
                                          input_features=[
                                              f_trip_distance, f_trip_time_duration],
                                          transform="f_trip_distance * f_trip_time_duration")

    f_trip_time_rounded = DerivedFeature(name="f_trip_time_rounded",
                                         feature_type=INT32,
                                         input_features=[f_trip_time_duration],
                                         transform="f_trip_time_duration % 10")

    location_id = TypedKey(key_column="DOLocationID",
                           key_column_type=ValueType.INT32,
                           description="location id in NYC",
                           full_name="nyc_taxi.location_id")
    agg_features = [Feature(name="f_location_avg_fare",
                            key=location_id,
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="cast_float(fare_amount)",
                                                              agg_func="AVG",
                                                              window="90d",
                                                              filter="fare_amount > 0"
                                                              )),
                    Feature(name="f_location_max_fare",
                            key=location_id,
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="cast_float(fare_amount)",
                                                              agg_func="MAX",
                                                              window="90d"))
                    ]

    agg_anchor = FeatureAnchor(name="aggregationFeatures",
                               source=sql_source,
                               features=agg_features)

    client.build_features(anchor_list=[agg_anchor, request_anchor], derived_feature_list=[
        f_trip_time_distance, f_trip_time_rounded])

    return client


def _sql_query_source():
    return SparkSqlSource(name="sparkSqlQuerySource", sql="SELECT * FROM green_tripdata_2020_04_with_index", event_timestamp_column="lpep_dropoff_datetime", timestamp_format="yyyy-MM-dd HH:mm:ss")


def _sql_table_source():
    return SparkSqlSource(name="sparkSqlTableSource", table="green_tripdata_2020_04_with_index", event_timestamp_column="lpep_dropoff_datetime", timestamp_format="yyyy-MM-dd HH:mm:ss")



import re
from copy import deepcopy
from typing import List, Optional, Union, Dict

from jinja2 import Template

from feathr.definition.dtype import FeatureType
from feathr.definition.transformation import ExpressionTransformation, Transformation, WindowAggTransformation
from feathr.definition.typed_key import DUMMY_KEY, TypedKey
from feathr.definition.feathrconfig import HoconConvertible


class FeatureBase(HoconConvertible):
    """The base class for features
    It has a feature name, feature type, and a convenient transformation used to produce its feature value.

    Attributes:
        name: Unique name of the feature. Only alphabet, numbers, and '_' are allowed in the name.
                It can not start with numbers. Note that '.' is NOT ALLOWED!
        feature_type: the feature value type. e.g. INT32, FLOAT, etc. feathr.dtype
        key: The key of this feature. e.g. user_id.
        transform: A transformation used to produce its feature value. e.g. amount * 10
        registry_tags: A dict of (str, str) that you can pass to feature registry for better organization. For example, you can use {"deprecated": "true"} to indicate this feature is deprecated, etc.
    """
    def __init__(self,
                 name: str,
                 feature_type: FeatureType,
                 transform: Optional[Union[str, Transformation]] = None,
                 key: Optional[Union[TypedKey, List[TypedKey]]] = [DUMMY_KEY],
                 registry_tags: Optional[Dict[str, str]] = None,
                 ):
        FeatureBase.validate_feature_name(name)

        # Validate the feature type
        if not isinstance(feature_type, FeatureType):
            raise KeyError(f'Feature type must be a FeatureType class, like INT32, but got {feature_type}')

        self.name = name
        self.feature_type = feature_type
        self.registry_tags=registry_tags
        self.key = key if isinstance(key, List) else [key]
        # feature_alias: Rename the derived feature to `feature_alias`. Default to feature name.
        self.feature_alias = name
        # If no transformation is specified, default to referencing the a field with the same name
        if transform is None:
            self.transform = ExpressionTransformation(name)
        elif isinstance(transform, str):
            self.transform = ExpressionTransformation(transform)
        else:
            self.transform = transform
        # An alias for the key in this feature. Default to its key column alias. Useful in derived features.
        # self.key could be null, when getting features from registry.
        self.key_alias = [k.key_column_alias for k in self.key if k]

    @classmethod
    def validate_feature_name(cls, feature_name: str) -> bool:
        """
        Only alphabet, numbers, and '_' are allowed in the name.
        It can not start with numbers. Note that '.' is NOT ALLOWED!
        This is because some compute engines, such as Spark, will consider them as operators in feature name.
        """
        if not feature_name:
            raise Exception('Feature name rule violation: empty feature name detected')

        feature_validator = re.compile(r"""^              # from the start of the string
                                           [a-zA-Z_]{1}   # first character can only be a letter or underscore 
                                           [a-zA-Z0-9_]+  # as many letters, numbers, or underscores as you like  
                                           $""",          # to the end of the string
                                       re.X)

        if not feature_validator.match(feature_name):
            raise Exception(
                'Feature name rule violation: only letters, numbers, and underscores are allowed in the name, ' +
                f'and the name cannot start with a number. name={feature_name}')

        return True

    def with_key(self, key_alias: Union[str, List[str]]):
        """Rename the feature key with the alias. This is useful in derived features that depends on
        the same feature with different keys."""
        cleaned_key_alias = [key_alias] if isinstance(key_alias, str) else key_alias
        assert(len(cleaned_key_alias) == len(self.key))
        new_key = []
        for i in range(0, len(cleaned_key_alias)):
            typed_key = deepcopy(self.key[i])
            typed_key.key_column_alias = cleaned_key_alias[i]
            new_key.append(typed_key)

        res = deepcopy(self)
        res.key = new_key
        res.key_alias = cleaned_key_alias
        return res

    def as_feature(self, feature_alias):
        """Provide the feature a different alias, which can be used to reference the feature in transformation
        expression. This is useful in derived features that depends on the same feature with different keys."""
        new_feature = deepcopy(self)
        new_feature.feature_alias = feature_alias
        return new_feature


class Feature(FeatureBase):
    """A feature is an individual measurable property or characteristic of an entity.
    It has a feature name, feature type, and a convenient row transformation used to produce its feature value.

    Attributes:
        name: Unique name of the feature. Only alphabet, numbers, and '_' are allowed in the name.
                It can not start with numbers. Note that '.' is NOT ALLOWED!
        feature_type: the feature value type. e.g. INT32, FLOAT, etc. Should be part of `feathr.dtype`
        key: The key of this feature. e.g. user_id.
        transform: A row transformation used to produce its feature value. e.g. amount * 10
        registry_tags: A dict of (str, str) that you can pass to feature registry for better organization. For example, you can use {"deprecated": "true"} to indicate this feature is deprecated, etc.
    """
    def __init__(self,
                name: str,
                feature_type: FeatureType,
                key: Optional[Union[TypedKey, List[TypedKey]]] = [DUMMY_KEY],
                transform: Optional[Union[str, Transformation]] = None,
                registry_tags: Optional[Dict[str, str]] = None,
                ):
        super(Feature, self).__init__(name, feature_type, transform, key, registry_tags)


    def to_feature_config(self) -> str:
        tm = Template("""
            {{feature.name}}: {
                {{feature.transform.to_feature_config()}}
                {{feature.feature_type.to_feature_config()}} 
            }
        """)
        return tm.render(feature=self)




from feathr.definition.anchor import FeatureAnchor
from feathr.definition.source import Source
from feathr.definition.feature_derivations import DerivedFeature
from feathr.definition.feature import Feature
from feathr.definition.transformation import Transformation
from typing import Set


class RepoDefinitions:
    """A list of shareable Feathr objects defined in the project."""
    def __init__(self,
                sources: Set[Source],
                features: Set[Feature],
                transformations: Set[Transformation],
                feature_anchors: Set[FeatureAnchor],
                derived_features: Set[DerivedFeature]) -> None:
        self.sources = sources
        self.features = features
        self.transformations = transformations
        self.feature_anchors = feature_anchors
        self.derived_features = derived_features


from feathr import FeatureAnchor
from feathr import Feature
from feathr import HdfsSource
from feathr import BOOLEAN, INT32, FLOAT, ValueType
from feathr import INPUT_CONTEXT
from feathr import WindowAggTransformation
from feathr import TypedKey

def test_request_feature_anchor_to_config():
    features = [
        Feature(name="trip_distance", feature_type=FLOAT),
        Feature(name="f_is_long_trip_distance",
                feature_type=BOOLEAN,
                transform="cast_float(trip_distance)>30"),
        Feature(name="f_day_of_week",
                feature_type=INT32,
                transform="dayofweek(lpep_dropoff_datetime)")
    ]

    anchor = FeatureAnchor(name="request_features",
                           source=INPUT_CONTEXT,
                           features=features)
    expected_non_agg_feature_config = """
           request_features: {
               source: PASSTHROUGH
               key.sqlExpr: [NOT_NEEDED]
               features: {
                    trip_distance: {
                        def.sqlExpr: "trip_distance"
                        type: {
                            type: TENSOR
                            tensorCategory: DENSE
                            dimensionType: []
                            valType: FLOAT
                        }
                    } 
                    f_is_long_trip_distance: {
                        def.sqlExpr: "cast_float(trip_distance)>30"
                        type: {
                            type: TENSOR
                            tensorCategory: DENSE
                            dimensionType: []
                            valType: BOOLEAN
                        } 
                    } 
                    f_day_of_week: { 
                        def.sqlExpr:"dayofweek(lpep_dropoff_datetime)"
                        type: {
                            type: TENSOR
                            tensorCategory: DENSE
                            dimensionType: []
                            valType: INT
                        }
                    }
               }
           }
    """
    assert ''.join(anchor.to_feature_config().split()) == ''.join(expected_non_agg_feature_config.split())


def test_non_agg_feature_anchor_to_config():
    batch_source = HdfsSource(name="nycTaxiBatchSource",
                              path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv",
                              event_timestamp_column="lpep_dropoff_datetime",
                              timestamp_format="yyyy-MM-dd HH:mm:ss")

    location_id = TypedKey(key_column="DOLocationID",
                     key_column_type=ValueType.INT32,
                     description="location id in NYC",
                     full_name="nyc_taxi.location_id")
    features = [
        Feature(name="f_loc_is_long_trip_distance",
                feature_type=BOOLEAN,
                transform="cast_float(trip_distance)>30", key=location_id),
        Feature(name="f_loc_day_of_week",
                feature_type=INT32,
                transform="dayofweek(lpep_dropoff_datetime)", key=location_id)
    ]

    anchor = FeatureAnchor(name="nonAggFeatures",
                           source=batch_source,
                           features=features)
    expected_non_agg_feature_config = """
           nonAggFeatures: {
               source: nycTaxiBatchSource
               key.sqlExpr: [DOLocationID]
               features: {
                    f_loc_is_long_trip_distance: {
                        def.sqlExpr: "cast_float(trip_distance)>30"
                        type: {
                            type: TENSOR
                            tensorCategory: DENSE
                            dimensionType: []
                            valType: BOOLEAN
                        } 
                    } 
                    f_loc_day_of_week: { 
                        def.sqlExpr:"dayofweek(lpep_dropoff_datetime)"
                        type: {
                            type: TENSOR
                            tensorCategory: DENSE
                            dimensionType: []
                            valType: INT
                        }
                    }
               }
           }
    """
    assert ''.join(anchor.to_feature_config().split()) == ''.join(expected_non_agg_feature_config.split())


def test_agg_anchor_to_config():
    batch_source = HdfsSource(name="nycTaxiBatchSource",
                              path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv",
                              event_timestamp_column="lpep_dropoff_datetime",
                              timestamp_format="yyyy-MM-dd HH:mm:ss")

    location_id = TypedKey(key_column="DOLocationID",
                     key_column_type=ValueType.INT32,
                     description="location id in NYC",
                     full_name="nyc_taxi.location_id")
    agg_features = [Feature(name="f_location_avg_fare",
                            key=location_id,
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="cast_float(fare_amount)",
                                                         agg_func="AVG",
                                                         window="90d")),
                    Feature(name="f_location_max_fare",
                            key=location_id,
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="cast_float(fare_amount)",
                                                         agg_func="MAX",
                                                         window="90d"))
                    ]

    agg_anchor = FeatureAnchor(name="aggregationFeatures",
                               source=batch_source,
                               features=agg_features)

    expected_agg_feature_config = """
            aggregationFeatures: {
                source: nycTaxiBatchSource
                key.sqlExpr: [DOLocationID]
                features: {
                    f_location_avg_fare: {
                        def: "cast_float(fare_amount)"
                        window: 90d
                        aggregation: AVG
                        type: {
                            type: TENSOR
                            tensorCategory: DENSE
                            dimensionType: []
                            valType: FLOAT
                        }
                    }
                    f_location_max_fare: {
                        def: "cast_float(fare_amount)"
                        window: 90d
                        aggregation: MAX
                        type: {
                            type: TENSOR
                            tensorCategory: DENSE
                            dimensionType: []
                            valType: FLOAT
                        }
                    }
                }
            }
        """
    assert ''.join(agg_anchor.to_feature_config().split()) == ''.join(expected_agg_feature_config.split())


from feathr.definition.materialization_settings import BackfillTime, MaterializationSettings
from feathr.definition.sink import RedisSink
import pytest
import glob
import shutil
import pandas as pd
import pandavro as pdx
from pathlib import Path
import os
from datetime import datetime, timedelta
from feathr import FeathrClient, ObservationSettings, FeatureQuery, TypedKey, HdfsSource, Feature, FeatureAnchor, INPUT_CONTEXT, FLOAT, INT32, BOOLEAN, DerivedFeature, WindowAggTransformation, ValueType
from feathr import BOOLEAN, FLOAT, INT32, ValueType

from test_utils.udfs import (add_new_dropoff_and_fare_amount_column, add_new_fare_amount)


def test_local_spark_get_offline_features():
    #This Test is for Local Spark only
    if not _is_local():
        return
    
    test_workspace_dir = Path(
        __file__).parent.resolve() / "test_user_workspace"

    client = _local_client_setup(test_workspace_dir)


    output_path, proc1 = _non_udf_features(client)
    client.wait_job_to_finish()

    df = parse_avro_result(output_path)
    assert df.shape[0] > 0

    shutil.rmtree('debug')
    return 

def test_local_spark_pyudf_get_offline_features():
    #This Test is for Local Spark only
    if not _is_local():
        return
    client = _local_client_setup()
    
    output_path, proc = _udf_features(client)

    client.wait_job_to_finish()
    df = parse_avro_result(output_path)
    assert df.shape[0] > 0
    shutil.rmtree('debug')
    return 

def test_local_spark_materialization():
    #This Test is for Local Spark only
    if not _is_local():
        return

    client: FeathrClient = _local_client_setup()
    results = _feature_gen_test(client)
    # just assume the job is successful without validating the actual result in Redis.
    shutil.rmtree('debug')
    return

def _is_local() -> bool:
    """
    to verify if test is running on local spark mode
    """
    if os.environ.get('SPARK_CONFIG__SPARK_CLUSTER') == 'local':
        return True
    else:
        return False

def _local_client_setup(local_workspace:str = None):
    if not local_workspace:
        local_workspace = Path(
        __file__).parent.resolve() / "test_user_workspace"

    os.chdir(local_workspace)
    client = FeathrClient(os.path.join(local_workspace, "feathr_config_local.yaml"), local_workspace_dir=local_workspace)
    return client

def _non_udf_features(client:FeathrClient = None):
    if not client:
        client = _local_client_setup()
    batch_source = HdfsSource(name="nycTaxiBatchSource",
                              path="./green_tripdata_2020-04_with_index.csv",
                              event_timestamp_column="lpep_dropoff_datetime",
                              timestamp_format="yyyy-MM-dd HH:mm:ss")

    f_trip_distance = Feature(name="f_trip_distance",
                              feature_type=FLOAT, transform="trip_distance")
    f_trip_time_duration = Feature(name="f_trip_time_duration",
                                   feature_type=INT32,
                                   transform="(to_unix_timestamp(lpep_dropoff_datetime) - to_unix_timestamp(lpep_pickup_datetime))/60")

    features = [
        f_trip_distance,
        f_trip_time_duration,
        Feature(name="f_is_long_trip_distance",
                feature_type=BOOLEAN,
                transform="cast_float(trip_distance)>30"),
        Feature(name="f_day_of_week",
                feature_type=INT32,
                transform="dayofweek(lpep_dropoff_datetime)"),
    ]


    request_anchor = FeatureAnchor(name="request_features",
                                   source=INPUT_CONTEXT,
                                   features=features)

    f_trip_time_distance = DerivedFeature(name="f_trip_time_distance",
                                          feature_type=FLOAT,
                                          input_features=[
                                              f_trip_distance, f_trip_time_duration],
                                          transform="f_trip_distance * f_trip_time_duration")

    f_trip_time_rounded = DerivedFeature(name="f_trip_time_rounded",
                                         feature_type=INT32,
                                         input_features=[f_trip_time_duration],
                                         transform="f_trip_time_duration % 10")

    location_id = TypedKey(key_column="DOLocationID",
                           key_column_type=ValueType.INT32,
                           description="location id in NYC",
                           full_name="nyc_taxi.location_id")
    agg_features = [Feature(name="f_location_avg_fare",
                            key=location_id,
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="cast_float(fare_amount)",
                                                              agg_func="AVG",
                                                              window="90d",
                                                              )),
                    Feature(name="f_location_max_fare",
                            key=location_id,
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="cast_float(fare_amount)",
                                                              agg_func="MAX",
                                                              window="90d"))
                    ]

    agg_anchor = FeatureAnchor(name="aggregationFeatures",
                               source=batch_source,
                               features=agg_features)

    client.build_features(anchor_list=[agg_anchor, request_anchor], derived_feature_list=[
        f_trip_time_distance, f_trip_time_rounded])

    location_id = TypedKey(key_column="DOLocationID",
                               key_column_type=ValueType.INT32,
                               description="location id in NYC",
                               full_name="nyc_taxi.location_id")

    feature_query = FeatureQuery(
            feature_list=["f_location_avg_fare"], key=location_id)
    
    settings = ObservationSettings(
        observation_path="./green_tripdata_2020-04_with_index.csv",
        #observation_path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv",
        event_timestamp_column="lpep_dropoff_datetime",
        timestamp_format="yyyy-MM-dd HH:mm:ss")

    now = datetime.now().strftime("%Y%m%d%H%M%S")
    output_path = os.path.join("debug", f"test_output_{now}")
    proc = client.get_offline_features(observation_settings=settings,
                            feature_query=feature_query,
                            output_path=output_path,
                            config_file_name = "feature_join_conf/feature_join_local.conf",
                            verbose=False)
    return output_path, proc

def _udf_features(client:FeathrClient = None):
    if not client:
        client = _local_client_setup()
    
    batch_source1 = HdfsSource(name="nycTaxiBatchSource_add_new_dropoff_and_fare_amount_column",
                              path="./green_tripdata_2020-04_with_index.csv",
                              preprocessing=add_new_dropoff_and_fare_amount_column,
                              event_timestamp_column="new_lpep_dropoff_datetime",
                              # event_timestamp_column="lpep_dropoff_datetime",
                              timestamp_format="yyyy-MM-dd HH:mm:ss")

    batch_source2 = HdfsSource(name="nycTaxiBatchSource_add_new_fare_amount",
                              path="./green_tripdata_2020-04_with_index.csv",
                              preprocessing=add_new_fare_amount,
                              event_timestamp_column="lpep_dropoff_datetime",
                              timestamp_format="yyyy-MM-dd HH:mm:ss")


    location_id = TypedKey(key_column="DOLocationID",
                           key_column_type=ValueType.INT32,
                           description="location id in NYC",
                           full_name="nyc_taxi.location_id")


    agg_features = [Feature(name="f_location_avg_fare",
                            key=location_id,
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="new_fare_amount",
                                                              agg_func="SUM",
                                                              window="90d")),
                    Feature(name="f_location_max_fare",
                            key=location_id,
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="new_fare_amount",
                                                              agg_func="MAX",
                                                              window="90d"))
                    ]

    agg_anchor = FeatureAnchor(name="aggregationFeatures",
                               source=batch_source1,
                               features=agg_features,
                               )

    pickup_time_as_id = TypedKey(key_column="lpep_pickup_datetime",
                                 key_column_type=ValueType.INT32,
                                 description="location id in NYC",
                                 full_name="nyc_taxi.location_id")

    features = [
        Feature(name="f_is_long_trip_distance",
                key=pickup_time_as_id,
                feature_type=FLOAT,
                transform="fare_amount_new"),
        Feature(name="f_day_of_week",
                key=pickup_time_as_id,
                feature_type=INT32,
                transform="dayofweek(lpep_dropoff_datetime)"),
    ]

    regular_anchor = FeatureAnchor(name="regular_anchor",
                                   source=batch_source2,
                                   features=features,
                                   )

    client.build_features(anchor_list=[agg_anchor, regular_anchor])

    feature_query = [FeatureQuery(
        feature_list=["f_is_long_trip_distance", "f_day_of_week"], key=pickup_time_as_id),
        FeatureQuery(
            feature_list=["f_location_avg_fare", "f_location_max_fare"], key=location_id)
    ]

    settings = ObservationSettings(
        observation_path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv",
        event_timestamp_column="lpep_dropoff_datetime",
        timestamp_format="yyyy-MM-dd HH:mm:ss")

    now = datetime.now().strftime("%Y%m%d%H%M%S")
    output_path = os.path.join("debug", f"test_output_{now}")
    proc = client.get_offline_features(observation_settings=settings,
                                feature_query=feature_query,
                                output_path=output_path)

    return output_path, proc

def _feature_gen_test(client:FeathrClient = None):
    if not client:
        client = _local_client_setup()
    batch_source = HdfsSource(name="nycTaxiBatchSource",
                              path="./green_tripdata_2020-04_with_index.csv",
                              event_timestamp_column="lpep_dropoff_datetime",
                              timestamp_format="yyyy-MM-dd HH:mm:ss")

    f_trip_distance = Feature(name="f_trip_distance",
                              feature_type=FLOAT, transform="trip_distance")
    f_trip_time_duration = Feature(name="f_trip_time_duration",
                                   feature_type=INT32,
                                   transform="(to_unix_timestamp(lpep_dropoff_datetime) - to_unix_timestamp(lpep_pickup_datetime))/60")

    features = [
        f_trip_distance,
        f_trip_time_duration,
        Feature(name="f_is_long_trip_distance",
                feature_type=BOOLEAN,
                transform="cast_float(trip_distance)>30"),
        Feature(name="f_day_of_week",
                feature_type=INT32,
                transform="dayofweek(lpep_dropoff_datetime)"),
    ]


    request_anchor = FeatureAnchor(name="request_features",
                                   source=INPUT_CONTEXT,
                                   features=features)

    f_trip_time_distance = DerivedFeature(name="f_trip_time_distance",
                                          feature_type=FLOAT,
                                          input_features=[
                                              f_trip_distance, f_trip_time_duration],
                                          transform="f_trip_distance * f_trip_time_duration")

    f_trip_time_rounded = DerivedFeature(name="f_trip_time_rounded",
                                         feature_type=INT32,
                                         input_features=[f_trip_time_duration],
                                         transform="f_trip_time_duration % 10")

    location_id = TypedKey(key_column="DOLocationID",
                           key_column_type=ValueType.INT32,
                           description="location id in NYC",
                           full_name="nyc_taxi.location_id")
    agg_features = [Feature(name="f_location_avg_fare",
                            key=location_id,
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="cast_float(fare_amount)",
                                                              agg_func="AVG",
                                                              window="90d",
                                                              )),
                    Feature(name="f_location_max_fare",
                            key=location_id,
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="cast_float(fare_amount)",
                                                              agg_func="MAX",
                                                              window="90d"))
                    ]

    agg_anchor = FeatureAnchor(name="aggregationFeatures",
                               source=batch_source,
                               features=agg_features)

    client.build_features(anchor_list=[agg_anchor, request_anchor], derived_feature_list=[
        f_trip_time_distance, f_trip_time_rounded])
    
    online_test_table = "localSparkTest"

    backfill_time = BackfillTime(start=datetime(
        2020, 4, 1), end=datetime(2020, 4, 2), step=timedelta(days=1))
    redisSink = RedisSink(table_name=online_test_table)
    settings = MaterializationSettings("LocalSparkTest",
                                       sinks=[redisSink],
                                       feature_names=[
                                           "f_location_avg_fare", "f_location_max_fare"],
                                       backfill_time=backfill_time)
    results = client.materialize_features(settings)

    client.wait_job_to_finish()
    res = client.get_online_features(online_test_table, '243', [
                                     'f_location_avg_fare', 'f_location_max_fare'])
    return res

def parse_avro_result(output_path:str):
    dataframe_list = []
    # assuming the result are in avro format
    for file in glob.glob(os.path.join(output_path, '*.avro')):
        dataframe_list.append(pdx.read_avro(file))
    
    vertical_concat_df = pd.concat(dataframe_list, axis=0)
    return vertical_concat_df


from typing import Optional
from jinja2 import Template
from loguru import logger
from feathr.definition.feathrconfig import HoconConvertible

class ConflictsAutoCorrection():
    """Conflicts auto-correction handler settings. 
       Used in feature join when some conflicts exist 
       between feature names and observation dataset columns.
    
    Attributes:
        rename_features: rename feature names when solving conflicts.
                         Default by 'False' which means to rename observation dataset columns.
        suffix: customized suffix to be applied to conflicts names. Default by "1"
                eg. The conflicted column name 'field' will become 'field_1' if suffix is "1"
    """
    def __init__(self, rename_features: bool = False, suffix: str = "1") -> None:
        self.rename_features = rename_features
        self.suffix = suffix
    
    def to_feature_config(self) -> str:
        tm = Template("""
            {% if auto_correction.rename_features %}
            renameFeatures: True
            {% else %}
            renameFeatures: False
            {% endif %}
            suffix: {{auto_correction.suffix}}  
        """)
        return tm.render(auto_correction=self)

class ObservationSettings(HoconConvertible):
    """Time settings of the observation data. Used in feature join.

    Attributes:
        observation_path: path to the observation dataset, i.e. input dataset to get with features
        event_timestamp_column (Optional[str]): The timestamp field of your record. As sliding window aggregation feature assume each record in the source data should have a timestamp column.
        timestamp_format (Optional[str], optional): The format of the timestamp field. Defaults to "epoch". Possible values are:
            - `epoch` (seconds since epoch), for example `1647737463`
            - `epoch_millis` (milliseconds since epoch), for example `1647737517761`
            - Any date formats supported by [SimpleDateFormat](https://docs.oracle.com/javase/8/docs/api/java/text/SimpleDateFormat.html).
        file_format: format of the dataset file. Default as "csv"
        is_file_path: if the 'observation_path' is a path of file (instead of a directory). Default as 'True'
        conflicts_auto_correction: settings about auto-correct feature names conflicts. 
                                   Default as None which means do not enable it.                                  
    """
    def __init__(self,
                 observation_path: str,
                 event_timestamp_column: Optional[str] = None,
                 timestamp_format: str = "epoch",
                 conflicts_auto_correction: ConflictsAutoCorrection = None,
                 file_format: str = "csv",
                 is_file_path: bool = True) -> None:
        self.event_timestamp_column = event_timestamp_column
        self.timestamp_format = timestamp_format
        self.observation_path = observation_path
        if observation_path.startswith("http"):
            logger.warning("Your observation_path {} starts with http, which is not supported. Consider using paths starting with wasb[s]/abfs[s]/s3.", observation_path)
        self.file_format = file_format
        self.is_file_path = is_file_path
        self.conflicts_auto_correction = conflicts_auto_correction
        
    def to_feature_config(self) -> str:
        tm = Template("""
                {% if setting.event_timestamp_column is not none %}
                settings: {
                    joinTimeSettings: {
                        timestampColumn: {
                            def: "{{setting.event_timestamp_column}}"
                            format: "{{setting.timestamp_format}}"
                        }
                    }
                    {% if setting.conflicts_auto_correction is not none %}
                    conflictsAutoCorrectionSettings: {
                        {{setting.conflicts_auto_correction.to_feature_config()}}
                    }
                    {% endif %}
                }
                {% endif %}
                observationPath: "{{setting.observation_path}}"
            """)
        return tm.render(setting=self)





from abc import abstractmethod
import copy
from typing import Callable, Dict, List, Optional
from feathr.definition.feathrconfig import HoconConvertible

from jinja2 import Template
from loguru import logger
from urllib.parse import urlparse, parse_qs
import json


class SourceSchema(HoconConvertible):
    pass


class AvroJsonSchema(SourceSchema):
    """Avro schema written in Json string form."""

    def __init__(self, schemaStr: str):
        self.schemaStr = schemaStr

    def to_feature_config(self):
        """Convert the feature anchor definition into internal HOCON format."""
        tm = Template("""
        schema: {
            type = "avro"
            avroJson:{{avroJson}}
        }
        """)
        avroJson = json.dumps(self.schemaStr)
        msg = tm.render(schema=self, avroJson=avroJson)
        return msg


class Source(HoconConvertible):
    """External data source for feature. Typically a data 'table'.

    Attributes:
         name: name of the source. It's used to differentiate from other sources.
         event_timestamp_column: column name or field name of the event timestamp
         timestamp_format: the format of the event_timestamp_column, e.g. yyyy/MM/DD, or EPOCH
         registry_tags: A dict of (str, str) that you can pass to feature registry for better organization.
                        For example, you can use {"deprecated": "true"} to indicate this source is deprecated, etc.
    """

    def __init__(self,
                 name: str,
                 event_timestamp_column: Optional[str] = "0",
                 timestamp_format: Optional[str] = "epoch",
                 registry_tags: Optional[Dict[str, str]] = None,
                 ) -> None:
        self.name = name
        self.event_timestamp_column = event_timestamp_column
        self.timestamp_format = timestamp_format
        self.registry_tags = registry_tags

    def __eq__(self, other):
        """A source is equal to another if name is equal."""
        return self.name == other.name

    def __hash__(self):
        """A source can be identified with the name"""
        return hash(self.name)

    def __str__(self):
        return self.to_feature_config()
    
    @abstractmethod
    def to_argument(self):
        pass


class InputContext(Source):
    """'Passthrough' source, a.k.a. request feature source. Feature source data is from the observation data itself.
    For example, you have an observation data table t1 and another feature data table t2. Some of your feature data
    can be transformed from the observation data table t1 itself, like geo location, then you can define that feature
    on top of the InputContext.
    """
    __SOURCE_NAME = "PASSTHROUGH"

    def __init__(self) -> None:
        super().__init__(self.__SOURCE_NAME, None, None)

    def to_feature_config(self) -> str:
        return "source: " + self.name

    def to_argument(self):
        raise TypeError("InputContext cannot be used as observation source")


class HdfsSource(Source):
    """A data source(table) stored on HDFS-like file system. Data can be fetch through a POSIX style path.

    Attributes:
        name (str): name of the source
        path (str): The location of the source data.
        preprocessing (Optional[Callable]): A preprocessing python function that transforms the source data for further feature transformation.
        event_timestamp_column (Optional[str]): The timestamp field of your record. As sliding window aggregation feature assume each record in the source data should have a timestamp column.
        timestamp_format (Optional[str], optional): The format of the timestamp field. Defaults to "epoch". Possible values are:
                                                    - `epoch` (seconds since epoch), for example `1647737463`
                                                    - `epoch_millis` (milliseconds since epoch), for example `1647737517761`
                                                    - Any date formats supported by [SimpleDateFormat](https://docs.oracle.com/javase/8/docs/api/java/text/SimpleDateFormat.html).
        registry_tags: A dict of (str, str) that you can pass to feature registry for better organization. For example, you can use {"deprecated": "true"} to indicate this source is deprecated, etc.
        time_partition_pattern(Optional[str]): Format of the time partitioned feature data. e.g. yyyy/MM/DD. All formats defined in dateTimeFormatter are supported.
        config:
            timeSnapshotHdfsSource: 
            {  
                location: 
                {    
                    path: "/data/somePath/daily/"  
                }  
                timePartitionPattern: "yyyy/MM/dd" 
            }
        Given the above HDFS path: /data/somePath/daily, 
        then the expectation is that the following sub directorie(s) should exist:
        /data/somePath/daily/{yyyy}/{MM}/{dd}
        postfix_path(Optional[str]): postfix path followed by the 'time_partition_pattern'. Given above config, if we have 'postfix_path' defined all contents under paths of the pattern '{path}/{yyyy}/{MM}/{dd}/{postfix_path}' will be visited.
    """

    def __init__(self, name: str, path: str, preprocessing: Optional[Callable] = None, event_timestamp_column: Optional[str] = None, timestamp_format: Optional[str] = "epoch", registry_tags: Optional[Dict[str, str]] = None, time_partition_pattern: Optional[str] = None, postfix_path: Optional[str] = None) -> None:
        super().__init__(name, event_timestamp_column,
                         timestamp_format, registry_tags=registry_tags)
        self.path = path
        self.preprocessing = preprocessing
        self.time_partition_pattern = time_partition_pattern
        self.postfix_path = postfix_path
        if path.startswith("http"):
            logger.warning(
                "Your input path {} starts with http, which is not supported. Consider using paths starting with wasb[s]/abfs[s]/s3.", path)

    def to_feature_config(self) -> str:
        tm = Template("""  
            {{source.name}}: {
                location: {path: "{{source.path}}"}
                {% if source.time_partition_pattern %}
                timePartitionPattern: "{{source.time_partition_pattern}}"
                {% endif %}
                {% if source.postfix_path %}
                postfixPath: "{{source.postfix_path}}"
                {% endif %}
                {% if source.event_timestamp_column %}
                    timeWindowParameters: {
                        timestampColumn: "{{source.event_timestamp_column}}"
                        timestampColumnFormat: "{{source.timestamp_format}}"
                    }
                {% endif %}
            } 
        """)
        msg = tm.render(source=self)
        return msg

    def __str__(self):
        return str(self.preprocessing) + '\n' + self.to_feature_config()

    def to_argument(self):
        return self.path

class SnowflakeSource(Source):
    """
    A data source for Snowflake

    Attributes:
        name (str): name of the source
        database (str): Snowflake Database
        schema (str): Snowflake Schema
        dbtable (Optional[str]): Snowflake Table
        query (Optional[str]): Query instead of snowflake table
        Either one of query or dbtable must be specified but not both.
        preprocessing (Optional[Callable]): A preprocessing python function that transforms the source data for further feature transformation.
        event_timestamp_column (Optional[str]): The timestamp field of your record. As sliding window aggregation feature assume each record in the source data should have a timestamp column.
        timestamp_format (Optional[str], optional): The format of the timestamp field. Defaults to "epoch". Possible values are:
                                                    - `epoch` (seconds since epoch), for example `1647737463`
                                                    - `epoch_millis` (milliseconds since epoch), for example `1647737517761`
                                                    - Any date formats supported by [SimpleDateFormat](https://docs.oracle.com/javase/8/docs/api/java/text/SimpleDateFormat.html).
        registry_tags: A dict of (str, str) that you can pass to feature registry for better organization. For example, you can use {"deprecated": "true"} to indicate this source is deprecated, etc.
    """
    def __init__(self, name: str, database: str, schema: str, dbtable: Optional[str] = None, query: Optional[str] = None, preprocessing: Optional[Callable] = None, event_timestamp_column: Optional[str] = None, timestamp_format: Optional[str] = "epoch", registry_tags: Optional[Dict[str, str]] = None) -> None:
        super().__init__(name, event_timestamp_column,
                         timestamp_format, registry_tags=registry_tags)
        self.preprocessing=preprocessing
        if dbtable is not None and query is not None:
            raise RuntimeError("Both dbtable and query are specified. Can only specify one..")
        if dbtable is None and query is None:
            raise RuntimeError("One of dbtable or query must be specified..")
        if dbtable is not None:
            self.dbtable = dbtable
        if query is not None:
            self.query = query
        self.database = database
        self.schema = schema
        self.path = self._get_snowflake_path(dbtable, query)

    def _get_snowflake_path(self, dbtable: Optional[str] = None, query: Optional[str] = None) -> str:
        """
        Returns snowflake path for registry.
        """
        if dbtable:
            return f"snowflake://snowflake_account/?sfDatabase={self.database}&sfSchema={self.schema}&dbtable={dbtable}"
        else:
            return f"snowflake://snowflake_account/?sfDatabase={self.database}&sfSchema={self.schema}&query={query}"
    
    def parse_snowflake_path(url: str) -> Dict[str, str]:
        """
        Parses snowflake path into dictionary of components for registry.
        """
        parse_result = urlparse(url)
        parsed_queries = parse_qs(parse_result.query)
        updated_dict = {key: parsed_queries[key][0] for key in parsed_queries}
        return updated_dict
    
    def to_feature_config(self) -> str:
        tm = Template("""  
            {{source.name}}: {
                type: SNOWFLAKE
                location: {
                    type: "snowflake"
                    {% if source.dbtable is defined %}
                    dbtable: "{{source.dbtable}}"
                    {% endif %}
                    {% if source.query is defined %}
                    query: "{{source.query}}"
                    {% endif %}
                    database: "{{source.database}}"
                    schema: "{{source.schema}}"
                }
                {% if source.event_timestamp_column %}
                    timeWindowParameters: {
                        timestampColumn: "{{source.event_timestamp_column}}"
                        timestampColumnFormat: "{{source.timestamp_format}}"
                    }
                {% endif %}
            } 
        """)
        msg = tm.render(source=self)
        return msg

    def __str__(self):
        return str(self.preprocessing) + '\n' + self.to_feature_config()

    def to_argument(self):
        return self.path

class JdbcSource(Source):
    def __init__(self, name: str, url: str = "", dbtable: Optional[str] = None, query: Optional[str] = None, auth: Optional[str] = None, preprocessing: Optional[Callable] = None, event_timestamp_column: Optional[str] = None, timestamp_format: Optional[str] = "epoch", registry_tags: Optional[Dict[str, str]] = None) -> None:
        super().__init__(name, event_timestamp_column, timestamp_format, registry_tags)
        self.preprocessing = preprocessing
        self.url = url
        if dbtable is not None:
            self.dbtable = dbtable
        if query is not None:
            self.query = query
        if auth is not None:
            self.auth = auth.upper()
            if self.auth not in ["USERPASS", "TOKEN"]:
                raise ValueError(
                    "auth must be None or one of following values: ['userpass', 'token']")

    def get_required_properties(self):
        if not hasattr(self, "auth"):
            return []
        if self.auth == "USERPASS":
            return ["%s_USER" % self.name.upper(), "%s_PASSWORD" % self.name.upper()]
        elif self.auth == "TOKEN":
            return ["%s_TOKEN" % self.name.upper()]

    def to_feature_config(self) -> str:
        tm = Template("""  
            {{source.name}}: {
                location: {
                    type: "jdbc"
                    url: "{{source.url}}"
                    {% if source.dbtable is defined %}
                    dbtable: "{{source.dbtable}}"
                    {% endif %}
                    {% if source.query is defined %}
                    query: "{{source.query}}"
                    {% endif %}
                    {% if source.auth is defined %}
                        {% if source.auth == "USERPASS" %}
                    user: "${{ "{" }}{{source.name}}_USER{{ "}" }}"
                    password: "${{ "{" }}{{source.name}}_PASSWORD{{ "}" }}"
                        {% else %}
                    useToken: true
                    token: "${{ "{" }}{{source.name}}_TOKEN{{ "}" }}"
                        {% endif %}
                    {% else %}
                    anonymous: true
                    {% endif %}
                }
                {% if source.event_timestamp_column %}
                    timeWindowParameters: {
                        timestampColumn: "{{source.event_timestamp_column}}"
                        timestampColumnFormat: "{{source.timestamp_format}}"
                    }
                {% endif %}
            } 
        """)
        source = copy.copy(self)
        source.name = self.name.upper()
        msg = tm.render(source=source)
        return msg

    def __str__(self):
        return str(self.preprocessing) + '\n' + self.to_feature_config()

    def to_argument(self):
        d = {
            "type": "jdbc",
            "url": self.url,
        }
        if hasattr(self, "dbtable"):
            d["dbtable"] = self.dbtable
        if hasattr(self, "query"):
            d["query"] = self.query
        if hasattr(self, "auth"):
            if self.auth == "USERPASS":
                d["user"] = "${" + self.name.upper() + "_USER}"
                d["password"] = "${" + self.name.upper() + "_PASSWORD}"
            elif self.auth == "TOKEN":
                d["useToken"] = True
                d["token"] = "${" + self.name.upper() + "_TOKEN}"
        else:
            d["anonymous"] = True
        return json.dumps(d)

class KafkaConfig:
    """Kafka config for a streaming source

    Attributes:
        brokers: broker/server address
        topics: Kafka topics
        schema: Kafka message schema
        """

    def __init__(self, brokers: List[str], topics: List[str], schema: SourceSchema):
        self.brokers = brokers
        self.topics = topics
        self.schema = schema


class KafKaSource(Source):
    """A kafka source object. Used in streaming feature ingestion."""

    def __init__(self, name: str, kafkaConfig: KafkaConfig):
        super().__init__(name)
        self.config = kafkaConfig

    def to_feature_config(self) -> str:
        tm = Template("""
{{source.name}}: {
    type: KAFKA
    config: {
        brokers: [{{brokers}}]
        topics: [{{topics}}]
        {{source.config.schema.to_feature_config()}}
    }
}
        """)
        brokers = '"'+'","'.join(self.config.brokers)+'"'
        topics = ','.join(self.config.topics)
        msg = tm.render(source=self, brokers=brokers, topics=topics)
        return msg

    def to_argument(self):
        raise TypeError("KafKaSource cannot be used as observation source")

class SparkSqlSource(Source):
    def __init__(self, name: str, sql: Optional[str] = None, table: Optional[str] = None, preprocessing: Optional[Callable] = None, event_timestamp_column: Optional[str] = None, timestamp_format: Optional[str] = "epoch", registry_tags: Optional[Dict[str, str]] = None) -> None:
        """ SparkSqlSource can use either a sql query or a table name as the source for Feathr job.
        name: name of the source
        sql: sql query to use as the source, either sql or table must be specified
        table: table name to use as the source, either sql or table must be specified
        preprocessing (Optional[Callable]): A preprocessing python function that transforms the source data for further feature transformation.
        event_timestamp_column (Optional[str]): The timestamp field of your record. As sliding window aggregation feature assume each record in the source data should have a timestamp column.
        timestamp_format (Optional[str], optional): The format of the timestamp field. Defaults to "epoch". Possible values are:
                                                    - `epoch` (seconds since epoch), for example `1647737463`
                                                    - `epoch_millis` (milliseconds since epoch), for example `1647737517761`
                                                    - Any date formats supported by [SimpleDateFormat](https://docs.oracle.com/javase/8/docs/api/java/text/SimpleDateFormat.html).
        registry_tags: A dict of (str, str) that you can pass to feature registry for better organization. For example, you can use {"deprecated": "true"} to indicate this source is deprecated, etc.
        """
        super().__init__(name, event_timestamp_column,
                         timestamp_format, registry_tags=registry_tags)
        self.source_type = 'sparksql'
        if sql is None and table is None:
            raise ValueError("Either `sql` or `table` must be specified")
        if sql is not None and table is not None:
            raise ValueError("Only one of `sql` or `table` can be specified")
        if sql is not None:
            self.sql = sql
        if table is not None:
            self.table = table
        self.preprocessing = preprocessing

    def to_feature_config(self) -> str:
        tm = Template("""  
            {{source.name}}: {
                location: {
                    type: "sparksql"
                    {% if source.sql is defined %}
                    sql: "{{source.sql}}"
                    {% elif source.table is defined %}
                    table: "{{source.table}}"
                    {% endif %}
                }
                {% if source.event_timestamp_column %}
                timeWindowParameters: {
                    timestampColumn: "{{source.event_timestamp_column}}"
                    timestampColumnFormat: "{{source.timestamp_format}}"
                }
                {% endif %}
            } 
        """)
        msg = tm.render(source=self)
        return msg

    def get_required_properties(self):
        return []

    def to_dict(self) -> Dict[str, str]:
        ret = {}
        ret["type"] = "sparksql"
        if hasattr(self, "sql"):
            ret["sql"] = self.sql
        elif hasattr(self, "table"):
            ret["table"] = self.table
        return ret        
    
    def to_argument(self):
        """
        One-line JSON string, used by job submitter
        """
        return json.dumps(self.to_dict())
    

class GenericSource(Source):
    """
    This class is corresponding to 'GenericLocation' in Feathr core, but only be used as Source.
    The class is not meant to be used by user directly, user should use its subclasses like `CosmosDbSource`
    """
    def __init__(self, name: str, format: str, mode: Optional[str] = None, options: Dict[str, str] = {}, preprocessing: Optional[Callable] = None, event_timestamp_column: Optional[str] = None, timestamp_format: Optional[str] = "epoch", registry_tags: Optional[Dict[str, str]] = None) -> None:
        super().__init__(name, event_timestamp_column,
                         timestamp_format, registry_tags=registry_tags)
        self.source_type = 'generic'
        self.format = format
        self.mode = mode
        self.preprocessing = preprocessing
        # Feathr Core uses HOCON format config, `.` will disrupt parsing.
        # In Feathr Core, GenericLocation will replace `__` back to `.`
        self.options = dict([(key.replace(".", "__"), options[key]) for key in options])

    def to_feature_config(self) -> str:
        tm = Template("""  
            {{source.name}}: {
                location: {
                    type: "generic"
                    format: "{{source.format}}"
                    {% if source.mode is defined %}
                    mode: "{{source.mode}}"
                    {% endif %}
                    {% for key,value in source.options.items() %}
                    "{{key}}":"{{value}}"
                    {% endfor %}
                }
                {% if source.event_timestamp_column %}
                timeWindowParameters: {
                    timestampColumn: "{{source.event_timestamp_column}}"
                    timestampColumnFormat: "{{source.timestamp_format}}"
                }
                {% endif %}
            } 
        """)
        msg = tm.render(source=self)
        return msg

    def get_required_properties(self):
        ret = []
        for option in self.options:
            start = option.find("${")
            if start >= 0:
                end = option[start:].find("}")
                if end >= 0:
                    ret.append(option[start+2:start+end])
        return ret

    def to_dict(self) -> Dict[str, str]:
        ret = self.options.copy()
        ret["type"] = "generic"
        ret["format"] = self.format
        if self.mode:
            ret["mode"] = self.mode
        return ret        
    
    def to_argument(self):
        """
        One-line JSON string, used by job submitter
        """
        return json.dumps(self.to_dict())


class CosmosDbSource(GenericSource):
    """
    Use CosmosDb as the data source
    """
    def __init__(self,
                 name: str,
                 endpoint: str,
                 database: str,
                 container: str,
                 preprocessing: Optional[Callable] = None,
                 event_timestamp_column: Optional[str] = None,
                 timestamp_format: Optional[str] = "epoch",
                 registry_tags: Optional[Dict[str, str]] = None):
        options = {
            'spark.cosmos.accountEndpoint': endpoint,
            'spark.cosmos.accountKey': "${%s_KEY}" % name.upper(),
            'spark.cosmos.database': database,
            'spark.cosmos.container': container
        }
        super().__init__(name,
                         format='cosmos.oltp',
                         mode="APPEND",
                         options=options,
                         preprocessing=preprocessing,
                         event_timestamp_column=event_timestamp_column, timestamp_format=timestamp_format,
                         registry_tags=registry_tags)

class ElasticSearchSource(GenericSource):
    """
    Use ElasticSearch as the data source
    """
    def __init__(self,
                 name: str,
                 host: str,
                 index: str,
                 ssl: bool = True,
                 auth: bool = True,
                 preprocessing: Optional[Callable] = None,
                 event_timestamp_column: Optional[str] = None,
                 timestamp_format: Optional[str] = "epoch",
                 registry_tags: Optional[Dict[str, str]] = None):
        """
        name: The name of the sink.
        host: ElasticSearch node, can be `hostname` or `hostname:port`, default port is 9200.
        index: The index to read the data.
        ssl: Set to `True` to enable SSL.
        auth: Set to `True` to enable authentication, you need to provide username/password from environment or KeyVault.
        preprocessing/event_timestamp_column/timestamp_format/registry_tags: See `HdfsSource`
        """
        options = {
            'es.nodes': host,
            'es.ssl': str(ssl).lower(),
            'es.resource': index,
        }
        if auth:
            options["es.net.http.auth.user"] = "${%s_USER}" % name.upper(),
            options["es.net.http.auth.pass"] = "${%s_PASSWORD}" % name.upper(),
        super().__init__(name,
                         format='org.elasticsearch.spark.sql',
                         mode="APPEND",
                         options=options,
                         preprocessing=preprocessing,
                         event_timestamp_column=event_timestamp_column, timestamp_format=timestamp_format,
                         registry_tags=registry_tags)

INPUT_CONTEXT = InputContext()



import os
from datetime import datetime, timedelta
from pathlib import Path

from feathr import (BackfillTime, MaterializationSettings)
from feathr import FeatureQuery
from feathr import ObservationSettings
from feathr import RedisSink
from feathr import TypedKey
from feathr import ValueType
from feathr.utils.job_utils import get_result_df
from test_fixture import (snowflake_test_setup, get_online_test_table_name)
from test_utils.constants import Constants


def test_feathr_online_store_agg_features():
    """
    Test FeathrClient() get_online_features and batch_get can get feature data correctly.
    """
    test_workspace_dir = Path(__file__).parent.resolve() / "test_user_workspace"

    client = snowflake_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))

    online_test_table = get_online_test_table_name("snowflakeSampleDemoFeature")

    backfill_time = BackfillTime(start=datetime(2020, 5, 20), end=datetime(2020, 5, 20), step=timedelta(days=1))
    redisSink = RedisSink(table_name= online_test_table)
    settings = MaterializationSettings(name="snowflakeSampleDemoFeature",
                                   sinks=[redisSink],
                                   feature_names=['f_snowflake_call_center_division_name',
                                                  'f_snowflake_call_center_zipcode'],
                                   backfill_time=backfill_time)
    client.materialize_features(settings, allow_materialize_non_agg_feature=True)
    # just assume the job is successful without validating the actual result in Redis. Might need to consolidate
    # this part with the test_feathr_online_store test case
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)

    res = client.get_online_features(online_test_table, '1',
                                     ['f_snowflake_call_center_division_name', 'f_snowflake_call_center_zipcode'])

    assert len(res) == 2
    assert res[0] != None
    assert res[1] != None
    res = client.multi_get_online_features(online_test_table,
                                    ['1', '2'],
                                    ['f_snowflake_call_center_division_name', 'f_snowflake_call_center_zipcode'])
    assert res['1'][0] != None
    assert res['1'][1] != None
    assert res['2'][0] != None
    assert res['2'][1] != None

    client._clean_test_data(online_test_table)

def test_feathr_get_offline_features():
    """
    Test get_offline_features() can get feature data from Snowflake source correctly.
    """
    test_workspace_dir = Path(__file__).parent.resolve() / "test_user_workspace"


    client = snowflake_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))
    call_sk_id = TypedKey(key_column="CC_CALL_CENTER_SK",
                          key_column_type=ValueType.INT32,
                          description="call center sk",
                          full_name="snowflake.CC_CALL_CENTER_SK")

    feature_query = FeatureQuery(
        feature_list=['f_snowflake_call_center_division_name', 'f_snowflake_call_center_zipcode'],
        key=call_sk_id)

    observation_path = client.get_snowflake_path(database="SNOWFLAKE_SAMPLE_DATA",schema="TPCDS_SF10TCL",dbtable="CALL_CENTER")
    settings = ObservationSettings(
        observation_path=observation_path)

    now = datetime.now()
     # set output folder based on different runtime
    if client.spark_runtime == 'databricks':
        output_path = ''.join(['dbfs:/feathrazure_cijob_snowflake','_', str(now.minute), '_', str(now.second), ".avro"])
    else:
        output_path = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/demo_data/snowflake_output','_', str(now.minute), '_', str(now.second), ".avro"])

    client.get_offline_features(observation_settings=settings,
                                feature_query=feature_query,
                                output_path=output_path)

    # assuming the job can successfully run; otherwise it will throw exception
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)

    res = get_result_df(client)
    # just assume there are results.
    assert res.shape[0] > 1

def test_client_get_snowflake_observation_path():
    """
    Test get_snowflake_path() returns correct snowflake observation path
    """
    test_workspace_dir = Path(__file__).parent.resolve() / "test_user_workspace"


    client = snowflake_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))
    snowflake_path_actual = client.get_snowflake_path(database="DATABASE", schema="SCHEMA", dbtable="TABLE")
    snowflake_path_expected = "snowflake://snowflake_account/?sfDatabase=DATABASE&sfSchema=SCHEMA&dbtable=TABLE"
    assert snowflake_path_actual == snowflake_path_expected



# NYC Taxi fare prediction sample dataset.
# Ref: https://www1.nyc.gov/site/tlc/about/tlc-trip-record-data.page
NYC_TAXI_SMALL_URL = (
    "https://azurefeathrstorage.blob.core.windows.net/public/sample_data/green_tripdata_2020-04_with_index.csv"
)

# Fraud detection sample datasets.
# Ref: https://github.com/microsoft/r-server-fraud-detection
FRAUD_DETECTION_ACCOUNT_INFO_URL = (
    "https://raw.github.com/microsoft/r-server-fraud-detection/master/Data/Account_Info.csv"
)

FRAUD_DETECTION_FRAUD_TRANSACTIONS_URL = (
    "https://raw.github.com/microsoft/r-server-fraud-detection/master/Data/Fraud_Transactions.csv"
)

FRAUD_DETECTION_UNTAGGED_TRANSACTIONS_URL = (
    "https://raw.github.com/microsoft/r-server-fraud-detection/master/Data/Untagged_Transactions.csv"
)

# Product recommendation sample datasets.
# Ref:
PRODUCT_RECOMMENDATION_USER_OBSERVATION_URL = (
    "https://azurefeathrstorage.blob.core.windows.net/public/sample_data/product_recommendation_sample/user_observation_mock_data.csv"
)

PRODUCT_RECOMMENDATION_USER_PROFILE_URL = (
    "https://azurefeathrstorage.blob.core.windows.net/public/sample_data/product_recommendation_sample/user_profile_mock_data.csv"
)

PRODUCT_RECOMMENDATION_USER_PURCHASE_HISTORY_URL = (
    "https://azurefeathrstorage.blob.core.windows.net/public/sample_data/product_recommendation_sample/user_purchase_history_mock_data.csv"
)

PRODUCT_RECOMMENDATION_PRODUCT_DETAIL_URL = (
    "https://azurefeathrstorage.blob.core.windows.net/public/sample_data/product_recommendation_sample/product_detail_mock_data.csv"
)

# Hotel review sample datasets.
# Ref: https://www.kaggle.com/datasets/datafiniti/hotel-reviews
HOTEL_REVIEWS_URL = (
    "https://raw.githubusercontent.com/Azure-Samples/azure-search-sample-data/main/hotelreviews/HotelReviews_data.csv"
)



from abc import ABC, abstractmethod

from typing import List, Tuple
from feathr.definition.feature_derivations import DerivedFeature
from feathr.definition.anchor import FeatureAnchor

class FeathrRegistry(ABC):
    """This is the abstract class for all the feature registries. All the feature registries should implement those interfaces.
    """

    @abstractmethod
    def register_features(self, anchor_list: List[FeatureAnchor] =[], derived_feature_list: List[DerivedFeature]=[]):
        """Registers features based on the current workspace

                Args:
                anchor_list: List of FeatureAnchors
                derived_feature_list: List of DerivedFeatures
        """
        pass


    @abstractmethod
    def list_registered_features(self, project_name: str) -> List[str]:
        """List all the already registered features under the given project.
        `project_name` must not be None or empty string because it violates the RBAC policy
        """
        pass

    @abstractmethod
    def list_dependent_entities(self, qualified_name: str):
        """
        Returns list of dependent entities for provided entity
        """
        pass

    @abstractmethod
    def delete_entity(self, qualified_name: str):
        """
        Deletes entity if it has no dependent entities
        """
        pass

    @abstractmethod
    def get_features_from_registry(self, project_name: str) -> Tuple[List[FeatureAnchor], List[DerivedFeature]]:
        """[Sync Features from registry to local workspace, given a project_name, will write project's features from registry to to user's local workspace]

        Args:
            project_name (str): project name.

        Returns:
            bool: Returns true if the job completed successfully, otherwise False
        """
        pass





from jinja2 import Template 
from feathr.definition.materialization_settings import MaterializationSettings


def _to_materialization_config(settings: MaterializationSettings):
    # produce materialization config
    tm = Template("""
            operational: {
            name: {{ settings.name }}
            endTime: "{{ settings.backfill_time.end.strftime('%Y-%m-%d %H:%M:%S') }}"
            endTimeFormat: "yyyy-MM-dd HH:mm:ss"
            resolution: {{ settings.resolution }}
            {% if settings.has_hdfs_sink == True %}
            enableIncremental = true
            {% endif %}
            output:[
                    {% for sink in settings.sinks %}
                        {{sink.to_feature_config()}}
                    {% endfor %}
                ]
            }
        features: [{{','.join(settings.feature_names)}}]
    """)
    msg = tm.render(settings=settings)
    return msg


from feathr import Aggregation
from feathr import Feature
from feathr import LookupFeature
from feathr import DerivedFeature
from feathr import FLOAT, FLOAT_VECTOR, ValueType, INT32_VECTOR
from feathr import TypedKey

def assert_config_equals(one, another):
    assert one.translate(str.maketrans('', '', ' \n\t\r')) == another.translate(str.maketrans('', '', ' \n\t\r'))
     
def test_single_key_lookup_feature_to_config():
    """Single key lookup feature config generation should work"""
    user_key = TypedKey(full_name="mockdata.user", key_column="user_id", key_column_type=ValueType.INT32, description="An user identifier")
    item_key = TypedKey(full_name="mockdata.item", key_column="item_id", key_column_type=ValueType.INT32, description="An item identifier")
    
    user_item = Feature(name="user_items", feature_type=INT32_VECTOR, key=user_key)
    item_price = Feature(name="item_price", feature_type=FLOAT_VECTOR, key=item_key)

    # A lookup feature
    lookup_feature = LookupFeature(name="user_avg_item_price",
                                    feature_type=FLOAT,
                                    key=user_key,
                                    base_feature=user_item,
                                    expansion_feature=item_price,
                                    aggregation=Aggregation.AVG)

    lookup_feature_config = """
    user_avg_item_price: {
        key: [user_id]
        join: {
            base: {key: [user_id], feature: user_items}
            expansion: {key: [item_id], feature: item_price}
        }
        aggregation: AVG
        type: {
            type: TENSOR
            tensorCategory: DENSE
            dimensionType: []
            valType: FLOAT
        }
    }"""
    assert_config_equals(lookup_feature.to_feature_config(), lookup_feature_config)
    assert(isinstance(lookup_feature, DerivedFeature))
 


from feathr import AvroJsonSchema
from feathr import KafKaSource
from feathr import KafkaConfig
from typing import List
import os
import random
from datetime import datetime, timedelta

from feathr import (BOOLEAN, FLOAT, INPUT_CONTEXT, INT32, STRING,
                    DerivedFeature, Feature, FeatureAnchor, HdfsSource,
                    TypedKey, ValueType, WindowAggTransformation, SnowflakeSource,
                    FeatureQuery,ObservationSettings)
from feathr import FeathrClient
from pyspark.sql import DataFrame


def basic_test_setup(config_path: str):

    now = datetime.now()
    # set workspace folder by time; make sure we don't have write conflict if there are many CI tests running
    os.environ['SPARK_CONFIG__DATABRICKS__WORK_DIR'] = ''.join(['dbfs:/feathrazure_cijob','_', str(now.minute), '_', str(now.second), '_', str(now.microsecond)]) 
    os.environ['SPARK_CONFIG__AZURE_SYNAPSE__WORKSPACE_DIR'] = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/feathr_github_ci','_', str(now.minute), '_', str(now.second) ,'_', str(now.microsecond)]) 
    
    client = FeathrClient(config_path=config_path)
    batch_source = HdfsSource(name="nycTaxiBatchSource",
                              path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv",
                              event_timestamp_column="lpep_dropoff_datetime",
                              timestamp_format="yyyy-MM-dd HH:mm:ss")

    f_trip_distance = Feature(name="f_trip_distance",
                              feature_type=FLOAT, transform="trip_distance")
    f_trip_time_duration = Feature(name="f_trip_time_duration",
                                   feature_type=INT32,
                                   transform="(to_unix_timestamp(lpep_dropoff_datetime) - to_unix_timestamp(lpep_pickup_datetime))/60")

    features = [
        f_trip_distance,
        f_trip_time_duration,
        Feature(name="f_is_long_trip_distance",
                feature_type=BOOLEAN,
                transform="cast_float(trip_distance)>30"),
        Feature(name="f_day_of_week",
                feature_type=INT32,
                transform="dayofweek(lpep_dropoff_datetime)"),
    ]


    request_anchor = FeatureAnchor(name="request_features",
                                   source=INPUT_CONTEXT,
                                   features=features)

    f_trip_time_distance = DerivedFeature(name="f_trip_time_distance",
                                          feature_type=FLOAT,
                                          input_features=[
                                              f_trip_distance, f_trip_time_duration],
                                          transform="f_trip_distance * f_trip_time_duration")

    f_trip_time_rounded = DerivedFeature(name="f_trip_time_rounded",
                                         feature_type=INT32,
                                         input_features=[f_trip_time_duration],
                                         transform="f_trip_time_duration % 10")

    location_id = TypedKey(key_column="DOLocationID",
                           key_column_type=ValueType.INT32,
                           description="location id in NYC",
                           full_name="nyc_taxi.location_id")
    agg_features = [Feature(name="f_location_avg_fare",
                            key=location_id,
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="cast_float(fare_amount)",
                                                              agg_func="AVG",
                                                              window="90d",
                                                              filter="fare_amount > 0"
                                                              )),
                    Feature(name="f_location_max_fare",
                            key=location_id,
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="cast_float(fare_amount)",
                                                              agg_func="MAX",
                                                              window="90d"))
                    ]

    agg_anchor = FeatureAnchor(name="aggregationFeatures",
                               source=batch_source,
                               features=agg_features)

    client.build_features(anchor_list=[agg_anchor, request_anchor], derived_feature_list=[
        f_trip_time_distance, f_trip_time_rounded])

    return client

def conflicts_auto_correction_setup(config_path: str):
    now = datetime.now()
    # set workspace folder by time; make sure we don't have write conflict if there are many CI tests running
    os.environ['SPARK_CONFIG__DATABRICKS__WORK_DIR'] = ''.join(['dbfs:/feathrazure_cijob','_', str(now.minute), '_', str(now.second), '_', str(now.microsecond)]) 
    os.environ['SPARK_CONFIG__AZURE_SYNAPSE__WORKSPACE_DIR'] = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/feathr_github_ci','_', str(now.minute), '_', str(now.second) ,'_', str(now.microsecond)]) 
    
    client = FeathrClient(config_path=config_path, local_workspace_dir="conflicts_test")
    batch_source = HdfsSource(name="nycTaxiBatchSource",
                          path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04_with_index.csv",
                          event_timestamp_column="lpep_dropoff_datetime",
                          timestamp_format="yyyy-MM-dd HH:mm:ss")
    location_id = TypedKey(key_column="DOLocationID",
                       key_column_type=ValueType.INT32,
                       description="location id in NYC",
                       full_name="nyc_taxi.location_id")
    pu_location_id = TypedKey(key_column="PULocationID",
                          key_column_type=ValueType.INT32,
                          description="location id in NYC",
                          full_name="nyc_taxi.location_id")

    agg_features = [Feature(name="tip_amount",
                        key=[location_id, pu_location_id],
                        feature_type=FLOAT,
                        transform=WindowAggTransformation(agg_expr="cast_float(fare_amount)",
                                                          agg_func="AVG",
                                                          window="3d")),
                Feature(name="total_amount",
                        key=[location_id, pu_location_id],
                        feature_type=FLOAT,
                        transform=WindowAggTransformation(agg_expr="cast_float(fare_amount)",
                                                          agg_func="MAX",
                                                          window="3d")),
                ]

    agg_anchor = FeatureAnchor(name="aggregationFeatures",
                           source=batch_source,
                           features=agg_features)
    
    client.build_features(anchor_list=[agg_anchor])

    return client

def composite_keys_test_setup(config_path: str):

    now = datetime.now()
    # set workspace folder by time; make sure we don't have write conflict if there are many CI tests running
    os.environ['SPARK_CONFIG__DATABRICKS__WORK_DIR'] = ''.join(['dbfs:/feathrazure_cijob','_', str(now.minute), '_', str(now.second), '_', str(now.microsecond)]) 
    os.environ['SPARK_CONFIG__AZURE_SYNAPSE__WORKSPACE_DIR'] = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/feathr_github_ci','_', str(now.minute), '_', str(now.second) ,'_', str(now.microsecond)]) 
    
    client = FeathrClient(config_path=config_path)
    batch_source = HdfsSource(name="nycTaxiBatchSource",
                              path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv",
                              event_timestamp_column="lpep_dropoff_datetime",
                              timestamp_format="yyyy-MM-dd HH:mm:ss")

    f_trip_distance = Feature(name="f_trip_distance",
                              feature_type=FLOAT, transform="trip_distance")
    f_trip_time_duration = Feature(name="f_trip_time_duration",
                                   feature_type=INT32,
                                   transform="(to_unix_timestamp(lpep_dropoff_datetime) - to_unix_timestamp(lpep_pickup_datetime))/60")

    features = [
        f_trip_distance,
        f_trip_time_duration,
        Feature(name="f_is_long_trip_distance",
                feature_type=BOOLEAN,
                transform="cast_float(trip_distance)>30"),
        Feature(name="f_day_of_week",
                feature_type=INT32,
                transform="dayofweek(lpep_dropoff_datetime)"),
    ]


    request_anchor = FeatureAnchor(name="request_features",
                                   source=INPUT_CONTEXT,
                                   features=features)

    f_trip_time_distance = DerivedFeature(name="f_trip_time_distance",
                                          feature_type=FLOAT,
                                          input_features=[
                                              f_trip_distance, f_trip_time_duration],
                                          transform="f_trip_distance * f_trip_time_duration")

    f_trip_time_rounded = DerivedFeature(name="f_trip_time_rounded",
                                         feature_type=INT32,
                                         input_features=[f_trip_time_duration],
                                         transform="f_trip_time_duration % 10")

    location_id = TypedKey(key_column="DOLocationID",
                           key_column_type=ValueType.INT32,
                           description="location id in NYC",
                           full_name="nyc_taxi.location_id")
    pu_location_id = TypedKey(key_column="PULocationID",
                          key_column_type=ValueType.INT32,
                          description="location id in NYC",
                          full_name="nyc_taxi.location_id")
    agg_features = [Feature(name="f_location_avg_fare",
                            key=[location_id,pu_location_id],
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="cast_float(fare_amount)",
                                                              agg_func="AVG",
                                                              window="90d",
                                                              filter="fare_amount > 0"
                                                              )),
                    Feature(name="f_location_max_fare",
                            key=[location_id,pu_location_id],
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="cast_float(fare_amount)",
                                                              agg_func="MAX",
                                                              window="90d"))
                    ]

    agg_anchor = FeatureAnchor(name="aggregationFeatures",
                               source=batch_source,
                               features=agg_features)

    client.build_features(anchor_list=[agg_anchor, request_anchor], derived_feature_list=[
        f_trip_time_distance, f_trip_time_rounded])

    return client


def snowflake_test_setup(config_path: str):
    now = datetime.now()
    # set workspace folder by time; make sure we don't have write conflict if there are many CI tests running
    os.environ['SPARK_CONFIG__DATABRICKS__WORK_DIR'] = ''.join(['dbfs:/feathrazure_cijob_snowflake','_', str(now.minute), '_', str(now.second), '_', str(now.microsecond)]) 
    os.environ['SPARK_CONFIG__AZURE_SYNAPSE__WORKSPACE_DIR'] = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/feathr_github_ci_snowflake','_', str(now.minute), '_', str(now.second), '_', str(now.microsecond)]) 

    client = FeathrClient(config_path=config_path)
    batch_source = SnowflakeSource(name="snowflakeSampleBatchSource",
                                   database="SNOWFLAKE_SAMPLE_DATA",
                                   schema="TPCDS_SF10TCL",
                                   dbtable="CALL_CENTER")
 
    call_sk_id = TypedKey(key_column="CC_CALL_CENTER_SK",
                          key_column_type=ValueType.INT32,
                          description="call center sk",
                          full_name="snowflake.CC_CALL_CENTER_SK")

    f_snowflake_call_center_division_name = Feature(name="f_snowflake_call_center_division_name",feature_type=STRING, transform="CC_DIVISION_NAME", key=call_sk_id)

    f_snowflake_call_center_zipcode = Feature(name="f_snowflake_call_center_zipcode",feature_type=STRING, transform="CC_ZIP", key=call_sk_id)



    features = [f_snowflake_call_center_division_name, f_snowflake_call_center_zipcode ]


    snowflakeFeatures = FeatureAnchor(name="snowflakeFeatures",
                                   source=batch_source,
                                   features=features)
    client.build_features(anchor_list=[snowflakeFeatures])
    return client

def kafka_test_setup(config_path: str):
    client = FeathrClient(config_path=config_path)
    schema = AvroJsonSchema(schemaStr="""
    {
        "type": "record",
        "name": "DriverTrips",
        "fields": [
            {"name": "driver_id", "type": "long"},
            {"name": "trips_today", "type": "int"},
            {
                "name": "datetime",
                "type": {"type": "long", "logicalType": "timestamp-micros"}
            }
        ]
    }
    """)
    stream_source = KafKaSource(name="kafkaStreamingSource",
                              kafkaConfig=KafkaConfig(brokers=["feathrazureci.servicebus.windows.net:9093"],
                                                      topics=["feathrcieventhub"],
                                                      schema=schema)
                              )

    driver_id = TypedKey(key_column="driver_id",
                          key_column_type=ValueType.INT64,
                          description="driver id",
                          full_name="nyc driver id")

    kafkaAnchor = FeatureAnchor(name="kafkaAnchor",
                                      source=stream_source,
                                      features=[Feature(name="f_modified_streaming_count",
                                                        feature_type=INT32,
                                                        transform="trips_today + 1",
                                                        key=driver_id),
                                                Feature(name="f_modified_streaming_count2",
                                                        feature_type=INT32,
                                                        transform="trips_today + randn() * cos(trips_today)", # make sure the supported SQL functions are tested
                                                        key=driver_id)]
                                      )
    client.build_features(anchor_list=[kafkaAnchor])
    return client

def registry_test_setup(config_path: str):
    # Use a new project name every time to make sure all features are registered correctly
    # Project name example: feathr_ci_registry_2022_09_24_01_02_30
    now = datetime.now()
    os.environ["project_config__project_name"] = f'feathr_ci_registry_{str(now)[:19].replace(" ", "_").replace(":", "_").replace("-", "_")}'

    client = FeathrClient(config_path=config_path, project_registry_tag={"for_test_purpose":"true"})
    request_anchor, agg_anchor, derived_feature_list = generate_entities()

    client.build_features(anchor_list=[agg_anchor, request_anchor], derived_feature_list=derived_feature_list)
    return client

def registry_test_setup_partially(config_path: str):
    """Register a partial of a project. Will call `generate_entities()` and register only the first anchor feature.
    """
    client = FeathrClient(config_path=config_path, project_registry_tag={"for_test_purpose":"true"})

    request_anchor, agg_anchor, derived_feature_list = generate_entities()
    agg_anchor.features = agg_anchor.features[:1]
    client.build_features(anchor_list=[agg_anchor, request_anchor], derived_feature_list=derived_feature_list)
    return client

def generate_entities():
    def add_new_dropoff_and_fare_amount_column(df: DataFrame):
        from pyspark.sql.functions import col
        df = df.withColumn("new_lpep_dropoff_datetime", col("lpep_dropoff_datetime"))
        df = df.withColumn("new_fare_amount", col("fare_amount") + 1000000)
        return df

    batch_source = HdfsSource(name="nycTaxiBatchSource",
                              path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv",
                              event_timestamp_column="lpep_dropoff_datetime",
                              timestamp_format="yyyy-MM-dd HH:mm:ss",
                              preprocessing=add_new_dropoff_and_fare_amount_column,
                              registry_tags={"for_test_purpose":"true"}
                              )

    f_trip_distance = Feature(name="f_trip_distance",
                              feature_type=FLOAT, transform="trip_distance",
                              registry_tags={"for_test_purpose":"true"}
                              )
    f_trip_time_duration = Feature(name="f_trip_time_duration",
                               feature_type=INT32,
                               transform="(to_unix_timestamp(lpep_dropoff_datetime) - to_unix_timestamp(lpep_pickup_datetime))/60")


    features = [
        f_trip_distance,
        f_trip_time_duration,
        Feature(name="f_is_long_trip_distance",
                feature_type=BOOLEAN,
                transform="cast_float(trip_distance)>30"),
        Feature(name="f_day_of_week",
                feature_type=INT32,
                transform="dayofweek(lpep_dropoff_datetime)"),
    ]


    request_anchor = FeatureAnchor(name="request_features",
                                   source=INPUT_CONTEXT,
                                   features=features,
                                   registry_tags={"for_test_purpose":"true"}
                                   )

    f_trip_time_distance = DerivedFeature(name="f_trip_time_distance",
                                          feature_type=FLOAT,
                                          input_features=[
                                              f_trip_distance, f_trip_time_duration],
                                          transform="f_trip_distance * f_trip_time_duration")

    f_trip_time_rounded = DerivedFeature(name="f_trip_time_rounded",
                                         feature_type=INT32,
                                         input_features=[f_trip_time_duration],
                                         transform="f_trip_time_duration % 10")
    f_trip_time_rounded_plus = DerivedFeature(name="f_trip_time_rounded_plus",
                                         feature_type=INT32,
                                         input_features=[f_trip_time_rounded],
                                         transform="f_trip_time_rounded + 100")

    location_id = TypedKey(key_column="DOLocationID",
                           key_column_type=ValueType.INT32,
                           description="location id in NYC",
                           full_name="nyc_taxi.location_id")
    pu_location_id = TypedKey(key_column="PULocationID",
                              key_column_type=ValueType.INT32,
                           full_name="nyc_taxi.pu_location_id"
                       )
    agg_features = [Feature(name="f_location_avg_fare",
                            key=[location_id,pu_location_id],
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="cast_float(fare_amount)",
                                                              agg_func="AVG",
                                                              window="90d"))
                    ]
    
    agg_anchor = FeatureAnchor(name="aggregationFeatures",
                               source=batch_source,
                               features=agg_features)
    
    derived_feature_list = [
                        f_trip_time_distance, f_trip_time_rounded, f_trip_time_rounded_plus]
    
    # shuffle the order to make sure they can be parsed correctly
    # Those input derived features can be in arbitrary order, but in order to parse the right dependencies, we need to reorder them internally in a certain order. 
    # This shuffle is to make sure that each time we have random shuffle for the input and make sure the internal sorting algorithm works (we are using topological sort).
    random.shuffle(derived_feature_list)
    return request_anchor,agg_anchor,derived_feature_list

def registry_test_setup_append(config_path: str):


    # use a new project name every time to make sure all features are registered correctly
    now = datetime.now()
    os.environ["project_config__project_name"] =  ''.join(['feathr_ci_registry','_', str(now.minute), '_', str(now.second), '_', str(now.microsecond)]) 

    client = FeathrClient(config_path=config_path, project_registry_tag={"for_test_purpose":"true"})

    def add_new_dropoff_and_fare_amount_column(df: DataFrame):
        from pyspark.sql.functions import col
        df = df.withColumn("new_lpep_dropoff_datetime", col("lpep_dropoff_datetime"))
        df = df.withColumn("new_fare_amount", col("fare_amount") + 1000000)
        return df

    batch_source = HdfsSource(name="nycTaxiBatchSource",
                              path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv",
                              event_timestamp_column="lpep_dropoff_datetime",
                              timestamp_format="yyyy-MM-dd HH:mm:ss",
                              preprocessing=add_new_dropoff_and_fare_amount_column,
                              registry_tags={"for_test_purpose":"true"}
                              )

    f_trip_distance = Feature(name="f_trip_distance",
                              feature_type=FLOAT, transform="trip_distance",
                              registry_tags={"for_test_purpose":"true"}
                              )
    f_trip_time_duration = Feature(name="f_trip_time_duration",
                               feature_type=INT32,
                               transform="(to_unix_timestamp(lpep_dropoff_datetime) - to_unix_timestamp(lpep_pickup_datetime))/60")


    features = [
        f_trip_distance,
        f_trip_time_duration,
        Feature(name="f_is_long_trip_distance",
                feature_type=BOOLEAN,
                transform="cast_float(trip_distance)>30"),
        Feature(name="f_day_of_week",
                feature_type=INT32,
                transform="dayofweek(lpep_dropoff_datetime)"),
    ]


    request_anchor = FeatureAnchor(name="request_features",
                                   source=INPUT_CONTEXT,
                                   features=features,
                                   registry_tags={"for_test_purpose":"true"}
                                   )

    f_trip_time_distance = DerivedFeature(name="f_trip_time_distance",
                                          feature_type=FLOAT,
                                          input_features=[
                                              f_trip_distance, f_trip_time_duration],
                                          transform="f_trip_distance * f_trip_time_duration")

    f_trip_time_rounded = DerivedFeature(name="f_trip_time_rounded",
                                         feature_type=INT32,
                                         input_features=[f_trip_time_duration],
                                         transform="f_trip_time_duration % 10")
    f_trip_time_rounded_plus = DerivedFeature(name="f_trip_time_rounded_plus",
                                         feature_type=INT32,
                                         input_features=[f_trip_time_rounded],
                                         transform="f_trip_time_rounded + 100")

    location_id = TypedKey(key_column="DOLocationID",
                           key_column_type=ValueType.INT32,
                           description="location id in NYC",
                           full_name="nyc_taxi.location_id")
    agg_features = [Feature(name="f_location_avg_fare",
                            key=location_id,
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="cast_float(fare_amount)",
                                                              agg_func="AVG",
                                                              window="90d")),
                    Feature(name="f_location_max_fare",
                            key=location_id,
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="cast_float(fare_amount)",
                                                              agg_func="MAX",
                                                              window="90d"))
                    ]

    agg_anchor = FeatureAnchor(name="aggregationFeatures",
                               source=batch_source,
                               features=agg_features)
    
    derived_feature_list = [
                        f_trip_time_distance, f_trip_time_rounded, f_trip_time_rounded_plus]
    
    # shuffle the order to make sure they can be parsed correctly
    # Those input derived features can be in arbitrary order, but in order to parse the right dependencies, we need to reorder them internally in a certain order. 
    # This shuffle is to make sure that each time we have random shuffle for the input and make sure the internal sorting algorithm works (we are using topological sort).
    random.shuffle(derived_feature_list)
    client.build_features(anchor_list=[agg_anchor, request_anchor], derived_feature_list=derived_feature_list)
    return client

def registry_test_setup_for_409(config_path: str, project_name: str):
    now = datetime.now()
    os.environ["project_config__project_name"] =  project_name

    client = FeathrClient(config_path=config_path, project_registry_tag={"for_test_purpose":"true"})

    # tranform in other sample is cast_float(trip_distance)>30
    # update this to trigger 409 conflict with the existing one
    features = [
        Feature(name="f_is_long_trip_distance",
                feature_type=BOOLEAN,
                transform="cast_float(trip_distance)>10"),
    ]

    request_anchor = FeatureAnchor(name="request_features",
                                   source=INPUT_CONTEXT,
                                   features=features,
                                   registry_tags={"for_test_purpose":"true"}
                                   )

    client.build_features(anchor_list=[request_anchor])
    return client

def get_online_test_table_name(table_name: str):
    # use different time for testing to avoid write conflicts
    now = datetime.now()
    res_table = '_'.join([table_name, str(now.minute), str(now.second)])
    print("The online Redis table is", res_table)
    return res_table

def time_partition_pattern_feature_gen_test_setup(config_path: str, data_source_path: str, local_workspace_dir: str = None, resolution: str = 'DAILY', postfix_path: str = ""):
    now = datetime.now()
    # set workspace folder by time; make sure we don't have write conflict if there are many CI tests running
    os.environ['SPARK_CONFIG__DATABRICKS__WORK_DIR'] = ''.join(['dbfs:/feathrazure_cijob','_', str(now.minute), '_', str(now.second), '_', str(now.microsecond)]) 
    os.environ['SPARK_CONFIG__AZURE_SYNAPSE__WORKSPACE_DIR'] = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/feathr_github_ci','_', str(now.minute), '_', str(now.second) ,'_', str(now.microsecond)]) 
    client = FeathrClient(config_path=config_path, local_workspace_dir=local_workspace_dir)

    if resolution == 'DAILY':
        if postfix_path != "":
            batch_source = HdfsSource(name="testTimePartitionSource",
                          path=data_source_path,
                          time_partition_pattern="yyyy/MM/dd",
                          postfix_path=postfix_path
                        )
        else:
            batch_source = HdfsSource(name="testTimePartitionSource",
                          path=data_source_path,
                          time_partition_pattern="yyyy/MM/dd"
                        )
    else:
        batch_source = HdfsSource(name="testTimePartitionSource",
                          path=data_source_path,
                          time_partition_pattern="yyyy/MM/dd/HH"
                        )
    key = TypedKey(key_column="key0",
               key_column_type=ValueType.INT32)
    agg_features = [
    Feature(name="f_loc_avg_output",
            key=[key],
            feature_type=FLOAT,
            transform=WindowAggTransformation(agg_expr="f_location_avg_fare",
                                              agg_func="AVG",
                                              window="3d")),
    Feature(name="f_loc_max_output",
            feature_type=FLOAT,
            key=[key],
            transform=WindowAggTransformation(agg_expr="f_location_max_fare",
                                              agg_func="MAX",
                                              window="3d")),
    ]

    agg_anchor = FeatureAnchor(name="testTimePartitionFeatures",
                           source=batch_source,
                           features=agg_features)
    client.build_features(anchor_list=[agg_anchor])
    return client

def time_partition_pattern_feature_join_test_setup(config_path: str, data_source_path: str, local_workspace_dir: str = None, resolution: str = 'DAILY', postfix_path: str = ""):
    now = datetime.now()
    # set workspace folder by time; make sure we don't have write conflict if there are many CI tests running
    os.environ['SPARK_CONFIG__DATABRICKS__WORK_DIR'] = ''.join(['dbfs:/feathrazure_cijob','_', str(now.minute), '_', str(now.second), '_', str(now.microsecond)]) 
    os.environ['SPARK_CONFIG__AZURE_SYNAPSE__WORKSPACE_DIR'] = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/feathr_github_ci','_', str(now.minute), '_', str(now.second) ,'_', str(now.microsecond)]) 
    client = FeathrClient(config_path=config_path, local_workspace_dir=local_workspace_dir)
    
    if postfix_path == "":
        if resolution == 'DAILY':
            batch_source_tpp = HdfsSource(name="nycTaxiBatchSource",
                          path=data_source_path,
                          time_partition_pattern="yyyy/MM/dd"
                        )
        else:
            batch_source_tpp = HdfsSource(name="nycTaxiBatchSource",
                          path=data_source_path,
                          time_partition_pattern="yyyy/MM/dd/HH"
                        )
    else:
        batch_source_tpp = HdfsSource(name="nycTaxiBatchSource",
                          path=data_source_path,
                          time_partition_pattern="yyyy/MM/dd",
                          postfix_path=postfix_path
                        )
    tpp_key = TypedKey(key_column="f_location_max_fare",
                    key_column_type=ValueType.FLOAT)
    tpp_features = [
    Feature(name="key0",
            key=tpp_key,
            feature_type=FLOAT,
            transform=WindowAggTransformation(agg_expr="key0",
                                              agg_func="LATEST",
                                              window="3d"
                    ))
    ]
    tpp_anchor = FeatureAnchor(name="tppFeatures",
                            source=batch_source_tpp,
                            features=tpp_features)
    client.build_features(anchor_list=[tpp_anchor])
    
    feature_query = FeatureQuery(feature_list=["key0"], key=tpp_key)
    settings = ObservationSettings(
        observation_path='wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/tpp_source.csv',
        event_timestamp_column="lpep_dropoff_datetime",
        timestamp_format="yyyy-MM-dd HH:mm:ss")
    return [client, feature_query, settings]


from copy import copy, deepcopy
from typing import List, Optional, Union, Dict

from jinja2 import Template

from feathr.definition.dtype import FeatureType
from feathr.definition.feature_derivations import DerivedFeature
from feathr.definition.feature import FeatureBase
from feathr.definition.transformation import RowTransformation
from feathr.definition.typed_key import DUMMY_KEY, TypedKey
from feathr.definition.aggregation import Aggregation

class LookupFeature(DerivedFeature):
    """A lookup feature is a feature defined on top of two other features, i.e. using the feature value of the base feature as key, to lookup the feature value from the expansion feature.
    e.g. a lookup feature user_purchased_item_avg_price could be key-ed by user_id, and computed by:
    base feature is user_purchased_item_ids. For a given user_id, it returns the item ids purchased by the user.
    expansion feature is item_price. For a given item id, it returns the item price.
    aggregation function is average of the item prices.

    Attributes:
        name: Derived feature name
        feature_type: Type of derived feature
        key: Join key of the derived feature
        base_feature: The feature value of this feature will be used as key to lookup from the expansion feature
        expansion_feature: The feature to be looked up
        aggregation: Specify the aggregation for the feature values lookup from the expansion feature, in the
                     case of the base feature value needed to be converted into multiple lookup keys,
                     e.g. feature value is an array and each value in the array is used once as a lookup key.
    """

    def __init__(self,
                name: str,
                feature_type: FeatureType,
                base_feature: FeatureBase,
                expansion_feature: FeatureBase,
                aggregation: Aggregation,
                key: Optional[Union[TypedKey, List[TypedKey]]] = [DUMMY_KEY],
                registry_tags: Optional[Dict[str, str]] = None,
                ):
        super(LookupFeature, self).__init__(name, feature_type, input_features=[base_feature, expansion_feature],
                                            transform="", key=key, registry_tags=registry_tags)
        self.base_feature = base_feature
        self.expansion_feature = expansion_feature
        self.aggregation = aggregation

    def to_feature_config(self) -> str:
        tm = Template("""
            {{lookup_feature.name}}: {
                key: [{{','.join(lookup_feature.key_alias)}}]
                join: {
                    base:{ 
                            key: [{{','.join(lookup_feature.base_feature.key_alias)}}],
                            feature: {{lookup_feature.base_feature.name}} 
                         }
                    expansion: { 
                            key: [{{','.join(lookup_feature.expansion_feature.key_alias)}}],
                            feature: {{lookup_feature.expansion_feature.name}}
                         } 
                }
                aggregation: {{lookup_feature.aggregation.name}}
                {{lookup_feature.feature_type.to_feature_config()}}
            }
        """)
        return tm.render(lookup_feature=self)



import string
import os
import pytest
from pathlib import Path

from feathr import FeatureBase
from feathr import (TypedKey, ValueType, FeatureQuery, ObservationSettings)
from feathr import FeathrClient

@pytest.mark.parametrize('bad_feature_name',
                         [None,
                          ''])
def test_feature_name_fails_on_empty_name(bad_feature_name: str):
    with pytest.raises(Exception, match="empty feature name"):
        FeatureBase.validate_feature_name(bad_feature_name)


def test_feature_name_fails_on_leading_number():
    with pytest.raises(Exception, match="cannot start with a number"):
        FeatureBase.validate_feature_name("4featurename")


def test_feature_name_fails_on_punctuation_chars():
    for char in set(string.punctuation) - set('_'):
        with pytest.raises(Exception,  match="only letters, numbers, and underscores are allowed"):
            FeatureBase.validate_feature_name(f"feature_{char}_name")


@pytest.mark.parametrize('feature_name',
                         ["feature_name",
                          "features4lyfe",
                          "f_4_feature_",
                          "_leading_underscores_are_ok",
                          "CapitalizedFeature"
                          ''])
def test_feature_name_validates_ok(feature_name: str):
    assert FeatureBase.validate_feature_name(feature_name)
    
def test_feature_name_conflicts_with_public_dataset_columns():
    test_workspace_dir = Path(
        __file__).parent.resolve() / "test_user_workspace"
    
    client = client = FeathrClient(os.path.join(test_workspace_dir, "feathr_config.yaml"))
    
    location_id = TypedKey(key_column="DOLocationID",
                            key_column_type=ValueType.INT32,
                            description="location id in NYC",
                            full_name="nyc_taxi.location_id")
    
    feature_query = FeatureQuery(
        feature_list=["trip_distance","fare_amount"], key=location_id)
    settings = ObservationSettings(
        observation_path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04_with_index.csv",
        event_timestamp_column="lpep_dropoff_datetime",
        timestamp_format="yyyy-MM-dd HH:mm:ss")
    output_path = "wasbs://fake_path"
    with pytest.raises(RuntimeError) as e:
        client.get_offline_features(observation_settings=settings,
                feature_query=feature_query,
                output_path=output_path
        )
    assert str(e.value) == "Feature names exist conflicts with dataset column names: trip_distance,fare_amount"
        
    settings = ObservationSettings(
        observation_path="wasbs://public@fake_file",
        event_timestamp_column="lpep_dropoff_datetime",
        timestamp_format="yyyy-MM-dd HH:mm:ss")
    output_path = "wasbs://fakepath"
    with pytest.raises(RuntimeError) as e:
        client.get_offline_features(observation_settings=settings,
                feature_query=feature_query,
                output_path=output_path,
                dataset_column_names=set(('trip_distance','fare_amount'))
        )
    assert str(e.value) == "Feature names exist conflicts with dataset column names: trip_distance,fare_amount"
    
def test_feature_name_conflicts_with_private_dataset_columns():
    test_workspace_dir = Path(
        __file__).parent.resolve() / "test_user_workspace"
    
    client = client = FeathrClient(os.path.join(test_workspace_dir, "feathr_config.yaml"))
    
    if client.spark_runtime == 'databricks':
        source_path = 'dbfs:/timePartitionPattern_test/df0/daily/2020/05/01/'
    else:
        source_path = 'abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/timePartitionPattern_test/df0/daily/2020/05/01/'
      
    location_id = TypedKey(key_column="DOLocationID",
                            key_column_type=ValueType.INT32,
                            description="location id in NYC",
                            full_name="nyc_taxi.location_id")
    
    feature_query = FeatureQuery(
        feature_list=["f_location_avg_fare","f_location_max_fare"], key=location_id)
    
    settings = ObservationSettings(
        observation_path=source_path,
        event_timestamp_column="lpep_dropoff_datetime",
        timestamp_format="yyyy-MM-dd HH:mm:ss",
        file_format="avro",
        is_file_path=False)
    output_path = "wasbs://fake_path"
    with pytest.raises(RuntimeError) as e:
        client.get_offline_features(observation_settings=settings,
                feature_query=feature_query,
                output_path=output_path
        )
    assert str(e.value) == "Feature names exist conflicts with dataset column names: f_location_avg_fare,f_location_max_fare"
        



import os
from pathlib import Path
import yaml

from loguru import logger

from azure.core.exceptions import ResourceNotFoundError
from feathr.secrets.akv_client import AzureKeyVaultClient

class EnvConfigReader(object):
    """A utility class to read Feathr environment variables either from os environment variables,
    the config yaml file or Azure Key Vault.
    If a key is set in the environment variable, ConfigReader will return the value of that environment variable.
    """
    akv_name: str = None      # Azure Key Vault name to use for retrieving config values.
    yaml_config: dict = None  # YAML config file content.

    def __init__(self, config_path: str):
        """Initialize the utility class.

        Args:
            config_path: Config file path.
        """
        if config_path is not None:
            config_path = Path(config_path)
            if config_path.is_file():
                try:
                    self.yaml_config = yaml.safe_load(config_path.read_text())
                except yaml.YAMLError as e:
                    logger.warning(e)

        self.akv_name = self.get("secrets__azure_key_vault__name")
        self.akv_client = AzureKeyVaultClient(self.akv_name) if self.akv_name else None

    def get(self, key: str, default: str = None) -> str:
        """Gets the Feathr config variable for the given key.
        It will retrieve the value in the following order:
            - From the environment variable if the key is set in the os environment variables.
            - From the config yaml file if the key exists.
            - From the Azure Key Vault.
        If the key is not found in any of the above, it will return `default`.

        Args:
            key: Config variable name. For example, `SPARK_CONFIG__DATABRICKS__WORKSPACE_INSTANCE_URL`
            default (optional): Default value to return if the key is not found. Defaults to None.

        Returns:
            Feathr client's config value.
        """
        val = self._get_variable_from_env(key)
        # `val` could be a boolean value, so we need to check if it is None.
        if val is None and self.yaml_config is not None:
            val = self._get_variable_from_file(key)
        if val is None and self.akv_name is not None:
            val = self._get_variable_from_akv(key)

        if val is not None:
            return val
        else:
            logger.info(f"Config {key} is not found in the environment variable, configuration file, or the remote key value store. Returning the default value: {default}.")
            return default

    def get_from_env_or_akv(self, key: str) -> str:
        """Gets the Feathr config variable for the given key.
        It will retrieve the value in the following order:
            - From the environment variable if the key is set in the os environment variables.
            - From the Azure Key Vault.
        If the key is not found in any of the above, it will return None.

        Args:
            key: Config variable name. For example, `ADLS_ACCOUNT`

        Returns:
            Feathr client's config value.
        """
        val = self._get_variable_from_env(key)
        # `val` could be a boolean value, so we need to check if it is None.
        if val is None and self.akv_name is not None:
            val = self._get_variable_from_akv(key)

        if val is not None:
            return val
        else:
            logger.warning(f"Config {key} is not found in the environment variable or the remote key value store.")
            return None

    def _get_variable_from_env(self, key: str) -> str:
        # make it work for lower case and upper case.
        conf_var = os.environ.get(key.lower(), os.environ.get(key.upper()))

        return conf_var

    def _get_variable_from_akv(self, key: str) -> str:
        try:
            # Azure Key Vault object name is case in-sensitive.
            # https://learn.microsoft.com/en-us/azure/key-vault/general/about-keys-secrets-certificates#vault-name-and-object-name
            return self.akv_client.get_feathr_akv_secret(key)
        except ResourceNotFoundError:
            logger.warning(f"Resource {self.akv_name} not found")

        return None

    def _get_variable_from_file(self, key: str) -> str:
        args = key.split("__")
        try:
            conf_var = self.yaml_config
            for arg in args:
                if conf_var is None:
                    break
                # make it work for lower case and upper case.
                conf_var = conf_var.get(arg.lower(), conf_var.get(arg.upper()))

            return conf_var
        except Exception as e:
            logger.warning(e)

        return None



import os
from datetime import datetime, timedelta
from pathlib import Path
from feathr.client import FeathrClient
from feathr.utils.job_utils import get_result_df

from pyspark.sql import DataFrame
from pyspark.sql.functions import col

from feathr import BOOLEAN, FLOAT, INT32, ValueType
from feathr import (BackfillTime, MaterializationSettings, FeatureQuery,
                    ObservationSettings, SparkExecutionConfiguration, ConflictsAutoCorrection)
from feathr import Feature
from feathr import FeatureAnchor
from feathr import INPUT_CONTEXT, HdfsSource
from feathr import RedisSink, HdfsSink, AerospikeSink
from feathr import TypedKey
from feathr.definition._materialization_utils import _to_materialization_config
from test_fixture import basic_test_setup, conflicts_auto_correction_setup
from test_fixture import get_online_test_table_name
from test_utils.constants import Constants
from logging import raiseExceptions
import pytest

def test_feature_materialization_config():
    backfill_time = BackfillTime(start=datetime(2020, 5, 20), end=datetime(2020, 5,20), step=timedelta(days=1))
    redisSink = RedisSink(table_name="nycTaxiDemoFeature")
    settings = MaterializationSettings("nycTaxiTable",
                                        sinks=[redisSink],
                                        feature_names=["f_location_avg_fare", "f_location_max_fare"],
                                        backfill_time=backfill_time)
    config = _to_materialization_config(settings)
    expected_config = """ 
        operational: {
            name: nycTaxiTable
            endTime: "2020-05-20 00:00:00"
            endTimeFormat: "yyyy-MM-dd HH:mm:ss"
            resolution: DAILY
            output:[
                {
                    name: REDIS
                    params: {
                        table_name: "nycTaxiDemoFeature"
                        features: [f_location_avg_fare,f_location_max_fare]
                    }
                }
            ]
        }
        features: [f_location_avg_fare, f_location_max_fare]
        """
    assert ''.join(config.split()) == ''.join(expected_config.split())

def test_feature_materialization_offline_config():
    backfill_time = BackfillTime(start=datetime(2020, 5, 20), end=datetime(2020, 5,20), step=timedelta(days=1))
    offlineSink = HdfsSink(output_path="abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/demo_data/output/hdfs_test.avro")
    settings = MaterializationSettings("nycTaxiTable",
                                       sinks=[offlineSink],
                                       feature_names=["f_location_avg_fare", "f_location_max_fare"],
                                       backfill_time=backfill_time)
    config = _to_materialization_config(settings)
    expected_config = """ 
        operational: {
            name: nycTaxiTable
            endTime: "2020-05-20 00:00:00"
            endTimeFormat: "yyyy-MM-dd HH:mm:ss"
            resolution: DAILY
            enableIncremental = true
            output:[
                {
                    name: HDFS
                    outputFormat: RAW_DATA                   
                    params: {
                        path: "abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/demo_data/output/hdfs_test.avro"
                        features: [f_location_avg_fare,f_location_max_fare]
                        storeName: "df0"
                    }

                }
            ]
        }
        features: [f_location_avg_fare, f_location_max_fare]
        """
    assert ''.join(config.split()) == ''.join(expected_config.split())

def test_feature_materialization_aerospike_sink_config():
    as_sink = AerospikeSink(name="aerospike",seedhost="20.57.186.153", port=3000, namespace="test", setname="test")
    backfill_time = BackfillTime(start=datetime(
        2020, 5, 20), end=datetime(2020, 5, 20), step=timedelta(days=1))
    settings = MaterializationSettings("nycTaxiTable",
                                       sinks=[as_sink],
                                       feature_names=[
                                           "avgfare", "maxfare"],
                                       backfill_time=backfill_time)
    os.environ[f"aerospike_USER"] = "feathruser"
    os.environ[f"aerospike_PASSWORD"] = "feathrpwd"
    expected_config = """
    operational: {
            name: nycTaxiTable
            endTime: "2020-05-20 00:00:00"
            endTimeFormat: "yyyy-MM-dd HH:mm:ss"
            resolution: DAILY
            output:[
                {
                    "name": "HDFS",                 
                    "params": {
                        "aerospike__seedhost": "20.57.186.153",
                        "aerospike__port": "3000",
                        "aerospike__namespace": "test",
                        "aerospike__user": "${AEROSPIKE_USER}",
                        "aerospike__password": "${AEROSPIKE_PASSWORD}",
                        "aerospike__set": "test",
                        "type": "generic",
                        "format": "aerospike",
                        "mode": "APPEND"
                    }

                }
            ]
        }
        features: [avgfare, maxfare]
    """
    config = _to_materialization_config(settings)
    assert ''.join(config.split()) == ''.join(expected_config.split())
    
def test_feature_materialization_daily_schedule():
    """Test back fill cutoff time for a daily range"""
    backfill_time = BackfillTime(start=datetime(2022, 3, 1), end=datetime(2022, 3, 5), step=timedelta(days=1))
    settings = MaterializationSettings("", [], [], backfill_time)
    expected = [datetime(2022, 3, day) for day in range(1, 6)]
    assert settings.get_backfill_cutoff_time() == expected
    
    

def test_feature_materialization_hourly_schedule():
    """Test back fill cutoff time for a hourly range"""
    backfill_time = BackfillTime(start=datetime(2022, 3, 1, 1), end=datetime(2022, 3, 1, 5), step=timedelta(hours=1))
    settings = MaterializationSettings("", [], [], backfill_time)
    expected = [datetime(2022,3, 1, hour) for hour in range(1, 6)]
    assert settings.get_backfill_cutoff_time() == expected


def test_feature_materialization_now_schedule():
    """Test back fill cutoff time without backfill."""
    settings = MaterializationSettings("", [], [])
    date = settings.get_backfill_cutoff_time()[0]
    expected = datetime.now()
    assert expected.year == date.year
    assert expected.month == date.month
    assert expected.day == date.day

def test_build_feature_verbose():
    """
    Test verbose for pretty printing features
    """
    test_workspace_dir = Path(__file__).parent.resolve() / "test_user_workspace"

    client = basic_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))

    # An anchor feature
    features = [
        Feature(name="trip_distance", feature_type=FLOAT),
        Feature(name="f_is_long_trip_distance",
                feature_type=BOOLEAN,
                transform="cast_float(trip_distance)>30"),
        Feature(name="f_day_of_week",
                feature_type=INT32,
                transform="dayofweek(lpep_dropoff_datetime)")
    ]

    anchor = FeatureAnchor(name="request_features",
                           source=INPUT_CONTEXT,
                           features=features)

    # Check pretty print
    client.build_features(anchor_list=[anchor], verbose=True)

def test_get_offline_features_verbose():
    """
    Test verbose for pretty printing feature query
    """

    test_workspace_dir = Path(__file__).parent.resolve() / "test_user_workspace"

    client = basic_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))

    location_id = TypedKey(key_column="DOLocationID",
                            key_column_type=ValueType.INT32)

    feature_query = FeatureQuery(feature_list=["f_location_avg_fare"], key=location_id)

    settings = ObservationSettings(
        observation_path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04",
        event_timestamp_column="lpep_dropoff_datetime",
        timestamp_format="yyyy-MM-dd HH:mm:ss"
    )

    now = datetime.now()

    # set output folder based on different runtime
    if client.spark_runtime == 'databricks':
        output_path = ''.join(['dbfs:/feathrazure_cijob','_', str(now.minute), '_', str(now.second), ".parquet"])
    else:
        output_path = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/demo_data/output','_', str(now.minute), '_', str(now.second), ".parquet"])

    # Check pretty print
    client.get_offline_features(
                                observation_settings=settings,
                                feature_query=feature_query,
                                output_path=output_path,
                                execution_configurations=SparkExecutionConfiguration({"spark.feathr.inputFormat": "parquet", "spark.feathr.outputFormat": "parquet"}),
                                verbose=True
                        )
    
def test_get_offline_features_auto_correct_dataset():
    test_workspace_dir = Path(__file__).parent.resolve() / "test_user_workspace"

    client = conflicts_auto_correction_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))
    
    now = datetime.now()
    location_id = TypedKey(key_column="DOLocationID",
                       key_column_type=ValueType.INT32,
                       description="location id in NYC",
                       full_name="nyc_taxi.location_id")
    feature_query = FeatureQuery(
    feature_list=["tip_amount", "total_amount"], key=location_id)
    settings = ObservationSettings(
        observation_path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04_with_index.csv",
        event_timestamp_column="lpep_dropoff_datetime",
        timestamp_format="yyyy-MM-dd HH:mm:ss",
        conflicts_auto_correction=ConflictsAutoCorrection(rename_features=False, suffix="test"))
     # set output folder based on different runtime
    if client.spark_runtime == 'databricks':
        output_path = ''.join(['dbfs:/feathrazure_cijob','_', str(now.minute), '_', str(now.second), ".avro"])
    else:
        output_path = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/demo_data/output','_', str(now.minute), '_', str(now.second), ".avro"])
    client.get_offline_features(observation_settings=settings,
                            feature_query=feature_query,
                            output_path=output_path
                           )
    client.wait_job_to_finish(timeout_sec=500)

    res_df = get_result_df(client, data_format="avro", res_url = output_path)
    assert res_df.shape[0] > 0
    
def test_get_offline_features_auto_correct_features():
    test_workspace_dir = Path(__file__).parent.resolve() / "test_user_workspace"

    client = conflicts_auto_correction_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))
    
    now = datetime.now()
    location_id = TypedKey(key_column="DOLocationID",
                       key_column_type=ValueType.INT32,
                       description="location id in NYC",
                       full_name="nyc_taxi.location_id")
    feature_query = FeatureQuery(
    feature_list=["tip_amount", "total_amount"], key=location_id)
    settings = ObservationSettings(
        observation_path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04_with_index.csv",
        event_timestamp_column="lpep_dropoff_datetime",
        timestamp_format="yyyy-MM-dd HH:mm:ss",
        conflicts_auto_correction=ConflictsAutoCorrection(rename_features=True, suffix="test"))
    # set output folder based on different runtime
    if client.spark_runtime == 'databricks':
        output_path = ''.join(['dbfs:/feathrazure_cijob','_', str(now.minute), '_', str(now.second), ".avro"])
    else:
        output_path = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/demo_data/output','_', str(now.minute), '_', str(now.second), ".avro"])
   
    client.get_offline_features(observation_settings=settings,
                            feature_query=feature_query,
                            output_path=output_path
                           )
    client.wait_job_to_finish(timeout_sec=500)

    res_df = get_result_df(client, data_format="avro", res_url = output_path)
    assert res_df.shape[0] > 0

def test_materialize_features_verbose():
    online_test_table = get_online_test_table_name("nycTaxiCITableMaterializeVerbose")
    test_workspace_dir = Path(__file__).parent.resolve() / "test_user_workspace"

    client: FeathrClient = basic_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))
    backfill_time = BackfillTime(start=datetime(2020, 5, 20), end=datetime(2020, 5, 20), step=timedelta(days=1))
    redisSink = RedisSink(table_name=online_test_table)
    settings = MaterializationSettings("nycTaxiTable",
                                       sinks=[redisSink],
                                       feature_names=[
                                           "f_location_avg_fare", "f_location_max_fare"],
                                       backfill_time=backfill_time)
    client.materialize_features(settings, verbose=True)
    client._clean_test_data(online_test_table)

def add_new_fare_amount(df: DataFrame) -> DataFrame:
    df = df.withColumn("fare_amount_new", col("fare_amount") + 8000000)

    return df

def test_delete_feature_from_redis():
    """
    Test FeathrClient() delete_feature_from_redis to remove feature from Redis.
    """

    test_workspace_dir = Path(__file__).parent.resolve() / "test_user_workspace"

    client: FeathrClient = basic_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))

    batch_source = HdfsSource(name="nycTaxiBatchSource_add_new_fare_amount",
                              path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv",
                              preprocessing=add_new_fare_amount,
                              event_timestamp_column="lpep_dropoff_datetime",
                              timestamp_format="yyyy-MM-dd HH:mm:ss")

    pickup_time_as_id = TypedKey(key_column="lpep_pickup_datetime",
                                 key_column_type=ValueType.INT32,
                                 description="location id in NYC",
                                 full_name="nyc_taxi.location_id")

    features = [
        Feature(name="f_is_long_trip_distance",
                key=pickup_time_as_id,
                feature_type=FLOAT,
                transform="fare_amount_new"),
        Feature(name="f_day_of_week",
                key=pickup_time_as_id,
                feature_type=INT32,
                transform="dayofweek(lpep_dropoff_datetime)"),
    ]

    regular_anchor = FeatureAnchor(name="request_features_add_new_fare_amount",
                                   source=batch_source,
                                   features=features,
                                   )

    client.build_features(anchor_list=[regular_anchor])

    online_test_table = get_online_test_table_name('nycTaxiCITableDeletion')

    backfill_time = BackfillTime(start=datetime(
        2020, 5, 20), end=datetime(2020, 5, 20), step=timedelta(days=1))
    redisSink = RedisSink(table_name=online_test_table)
    settings = MaterializationSettings(name="py_udf",
                                       sinks=[redisSink],
                                       feature_names=[
                                           "f_is_long_trip_distance",
                                           "f_day_of_week"
                                       ],
                                       backfill_time=backfill_time)
    client.materialize_features(settings, allow_materialize_non_agg_feature=True)
    
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)
    
    res = client.get_online_features(online_test_table, '2020-04-01 07:21:51', [
        'f_is_long_trip_distance', 'f_day_of_week'])

    assert len(res) == 2

    assert res[0] != None
    assert res[1] != None

    # Delete online feature stored in Redis
    client.delete_feature_from_redis(online_test_table, '2020-04-01 07:21:51', 'f_is_long_trip_distance')
    
    # Check if the online feature is deleted successfully
    res = client.get_online_features(online_test_table, '265', ['f_location_avg_fare'])

    assert len(res) == 1
    assert res[0] == None
    
    client._clean_test_data(online_test_table)

def test_feature_list_on_input_context():
    with pytest.raises(RuntimeError) as e_info:
        test_workspace_dir = Path(__file__).parent.resolve() / "test_user_workspace"

        client: FeathrClient = basic_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))
        online_test_table = get_online_test_table_name('nycTaxiCITableDeletion')
        redisSink = RedisSink(table_name=online_test_table)
        settings = MaterializationSettings(name="py_udf",
                                       sinks=[redisSink],
                                       feature_names=[
                                           "f_location_avg_fare",
                                           "f_day_of_week"
                                       ])
        client.materialize_features(settings, allow_materialize_non_agg_feature=True)
    assert e_info is not None
    assert e_info.value.args[0] == "Materializing features that are defined on INPUT_CONTEXT is not supported. f_day_of_week is defined on INPUT_CONTEXT so you should remove it from the feature list in MaterializationSettings."  


from feathr.anchor import FeatureAnchor
from feathr.feature import Feature
from feathr.dtype import BOOLEAN, INT32, ValueType
from feathr.typed_key import TypedKey
from feathr.source import HdfsSource

batch_source = HdfsSource(name="nycTaxiBatchSource",
                          path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv",
                          event_timestamp_column="lpep_dropoff_datetime",
                          timestamp_format="yyyy-MM-dd HH:mm:ss")

location_id = TypedKey(key_column="DOLocationID",
                       key_column_type=ValueType.INT32,
                       description="location id in NYC",
                       full_name="nyc_taxi.location_id")
features = [
    Feature(name="f_loc_is_long_trip_distance",
            feature_type=BOOLEAN,
            transform="cast_float(trip_distance)>30", key=location_id),
    Feature(name="f_loc_day_of_week",
            feature_type=INT32,
            transform="dayofweek(lpep_dropoff_datetime)", key=location_id)
]

anchor = FeatureAnchor(name="nonAggFeatures",
                       source=batch_source,
                       features=features)